from typing import (
    TYPE_CHECKING,
    cast,
    Tuple,
    Union,
    TypeVar,
    Dict,
    List,
    Optional,
    Sequence,
    Deque,
    Any,
    NamedTuple,
)
from abc import ABC, abstractmethod
import re
from json import JSONDecodeError
from copy import deepcopy
from enum import Enum
from collections import deque
from itertools import accumulate, chain
from functools import cached_property
from operator import add
from urllib.parse import unquote_plus
from flask.globals import current_app
from jinja2 import Undefined
from lxml import etree
from lxml.html import Element
from flask import json, jsonify, g
from werkzeug.exceptions import NotFound
from werkzeug import Response
from .common import (
    exec_name_to_full_name,
    is_direct_child_name,
    is_child_name,
    is_page_exec_name,
    parent_exec_name,
    load_param,
    # json_object_hook,
    # json_default,
)
from .exceptions import AccessDenied, Forbidden, JembeError, Unauthorized
from .component_config import ComponentConfig, RedisplayFlag as RedisplayFlag


if TYPE_CHECKING:  # pragma: no cover
    import jembe
    from flask import Request
    from jembe.component_config import ComponentListener


class Event:
    def __init__(
        self,
        source_exec_name: str,
        source: Optional["jembe.Component"],
        event_name: str,
        to: Optional[str],
        params: dict,
    ):
        self.source_exec_name = source_exec_name
        self.name = event_name
        self.params = params
        self.to: Optional[str] = to

        self.source: Optional["jembe.Component"] = source

    def __repr__(self):
        return "Event: source={}, name={}, to={} params={}".format(
            self.source_exec_name, self.name, self.to, self.params
        )

    # can access parameters like attribures if name does not colide
    def __getattr__(self, name):
        if name in self.params.keys():
            return self.params[name]
        return super().__getattribute__(name)

    def __setattr__(self, name, value):
        if "params" in self.__dict__ and name in self.params:
            self.params[name] = value
        return super().__setattr__(name, value)

    @cached_property
    def source_full_name(self) -> str:
        if self.source:
            return self.source._config.full_name
        else:
            return exec_name_to_full_name(self.source_exec_name)

    @cached_property
    def source_name(self) -> str:
        if self.source and self.source._config.name:
            return self.source._config.name
        else:
            return exec_name_to_full_name(self.source_full_name).split("/")[-1]


class SystemEvents(Enum):
    # after display action is executed
    DISPLAY = "_display"
    # after exception is raised by command
    EXCEPTION = "_exception"


TCommand = TypeVar("TCommand", bound="Command")


class Command(ABC):
    jembe: "jembe.Jembe"

    def __init__(self, component_exec_name: str):
        self.component_exec_name = component_exec_name
        self.processor: "Processor"
        self.is_mounted = False

    def mount(self: TCommand, processor: "Processor") -> TCommand:
        if self.is_mounted:
            raise JembeError("Command {} is already mounted".format(self.__repr__()))

        self.is_mounted = True
        self.processor = processor
        return self

    @abstractmethod
    def execute(self):
        raise NotImplementedError()

    def __repr__(self):
        return "Command: {} Base".format(self.component_exec_name)

    def get_before_emit_commands(self) -> Sequence["EmitCommand"]:
        return ()

    def get_after_emit_commands(self) -> Sequence["EmitCommand"]:
        return ()

    @property
    def component_full_name(self):
        return exec_name_to_full_name(self.component_exec_name)


class CallActionCommand(Command):
    def __init__(
        self,
        component_exec_name: str,
        action_name: str,
        args: Optional[Sequence[Any]] = None,
        kwargs: Optional[dict] = None,
    ):
        super().__init__(component_exec_name)
        self.action_name = action_name
        self.args = args if args is not None else tuple()
        self.kwargs = kwargs if kwargs is not None else dict()

        self._do_reinject_into_children = False
        self._component_state_before_execute = None

    def execute(self):
        component = self.processor.components[self.component_exec_name]
        cconfig = component._config
        if self.action_name not in cconfig.component_actions:
            raise JembeError(
                "Action {}.{} does not exist or is not marked as public action".format(
                    cconfig.full_name, self.action_name
                )
            )
        # check if action passes access control
        if not component.ac_check(self.action_name):
            raise ComponentConfig.DEFAULT_AC_EXCEPTION()

        # if component injects params into children save component state json hesh
        if (
            cconfig._inject_into_components is not None
            or cconfig.component_class._jembe_inject_into_overriden
        ):
            self._component_state_before_execute = component.state.tojsondict(
                component, True
            )

        # execute action
        action_result = getattr(component, self.action_name)(*self.args, **self.kwargs)
        component._jembe_has_action_or_listener_executed = True

        # process action result
        if action_result is None or (
            isinstance(action_result, bool) and action_result == True
        ):
            if self.component_exec_name in self.processor.components.keys():
                # after executing action that returns True or None
                # component should be rendered by executing display
                self.processor.add_command(
                    CallDisplayCommand(
                        self.component_exec_name, force=action_result == True
                    ),
                    end=True,
                )
        elif isinstance(action_result, bool) and action_result == False:
            # Do nothing
            pass
        elif isinstance(action_result, Response):
            # TODO return response and
            # stop fruther processiong of requst with jembe processor
            # TODO jembe.js should be able to handle direct reponses
            # TODO make the same for listener
            raise NotImplementedError()
        else:
            raise JembeError(
                "Invalid action result type: {}.{} {}".format(
                    cconfig.full_name,
                    self.action_name,
                    action_result,
                )
            )

    def get_before_emit_commands(self) -> Sequence["EmitCommand"]:
        return ()

    def get_after_emit_commands(self) -> Sequence["EmitCommand"]:
        commands: List["EmitCommand"] = []
        # reinject params to child components
        if self._component_state_before_execute is not None:
            component = self.processor.components[self.component_exec_name]
            current_state = self.processor.components[
                self.component_exec_name
            ].state.tojsondict(component, True)
            if current_state != self._component_state_before_execute:
                for exec_name, render in self.processor.renderers.items():
                    if render.fresh and is_direct_child_name(
                        self.component_exec_name, exec_name
                    ):
                        # reinitialise component is freshly rerendered forcing it to apply
                        # new injected params
                        commands.extend((InitialiseCommand(exec_name, dict()),))
        return commands

    def __repr__(self):
        return "CallAction({}, {})".format(self.component_exec_name, self.action_name)


class CallDisplayCommand(CallActionCommand):
    def __init__(
        self,
        component_exec_name: str,
        force: bool = False,
        displayed_by_exec_name: Optional[str] = None,
    ) -> None:
        super().__init__(
            component_exec_name=component_exec_name,
            action_name=ComponentConfig.DEFAULT_DISPLAY_ACTION,
            args=None,
            kwargs=None,
        )
        self.force = force
        self.displayed_by_exec_name = displayed_by_exec_name
        self._displayed_components: List[str] = []

    def mount(self, processor: "Processor") -> "CallDisplayCommand":
        if self.displayed_by_exec_name:
            if (
                processor._processing_command
                and isinstance(processor._processing_command, CallDisplayCommand)
                and processor._processing_command.component_exec_name
                == self.displayed_by_exec_name
            ):
                processor._processing_command._displayed_components.append(
                    self.component_exec_name
                )
        return super().mount(processor)

    @cached_property
    def _component(self) -> "jembe.Component":
        return self.processor.components[self.component_exec_name]

    @cached_property
    def _redisplay_needed(self):
        if self.component_exec_name in self.processor.renderers and not (
            self.force
            or (
                # redisplay component created without merging params even if
                # it is rendered with same state (in order to provoke chain redisplayint)
                not self._component._jembe_merged_existing_params
                and not self.processor.renderers[self.component_exec_name].fresh
            )
            or RedisplayFlag.WHEN_ON_PAGE in self._component._config.redisplay
            or RedisplayFlag.WHEN_DISPLAY_EXECUTED in self._component._config.redisplay
            or (
                RedisplayFlag.WHEN_STATE_CHANGED in self._component._config.redisplay
                and self.processor.renderers[self.component_exec_name].state_jsondict
                != self._component.state.tojsondict(self._component, True)
            )
        ):
            # if compoent is already displayed/rendered in same state and no force is set
            # no need to execute display again because it should return same result
            return False
        return True

    def execute(self) -> Optional["Response"]:
        if not self._redisplay_needed:
            return None

        # check if action passes access control
        if not self.force and not self._component.ac_check():
            raise ComponentConfig.DEFAULT_AC_EXCEPTION()

        # execute action
        action_result = getattr(self._component, self.action_name)(
            *self.args, **self.kwargs
        )
        # process action result
        if isinstance(action_result, str):
            # save component display responses in memory
            # Add component html to processor rendererd
            self.processor.renderers[self.component_exec_name] = ComponentRender(
                True,
                self._component.state.tojsondict(self._component, True),
                self._component._jembe_disabled_actions.copy(),
                self._component.state._injected_params_names,
                self._component.url,
                self._component._config.changes_url,
                action_result,
                self._displayed_components,
            )
        elif isinstance(action_result, Response):
            # TODO If self.component is component directly requested via http request
            # othervise raise JembeError
            # if not self.processor.renderers:
            #     return action_result
            # else:
            #     raise JembeError(
            #         "{} action should be first one to return responses if that response is not html string but full response object".format(
            #             self._component._config.full_name
            #         )
            #     )
            return action_result
        elif not isinstance(action_result, str):
            # Display should return html string if not raise JembeError
            raise JembeError(
                "{} action of {} shuld return html string not {}".format(
                    ComponentConfig.DEFAULT_DISPLAY_ACTION,
                    self._component._config.full_name,
                    action_result,
                )
            )
        else:
            raise JembeError(
                "Invalid display result type: {}.{} {}".format(
                    self._component._config.full_name,
                    self.action_name,
                    action_result,
                )
            )
        return None

    def get_before_emit_commands(self) -> Sequence["EmitCommand"]:
        return super().get_before_emit_commands()

    def get_after_emit_commands(self) -> Sequence["EmitCommand"]:
        commands = list(super().get_after_emit_commands())
        if self._redisplay_needed:
            commands.append(
                EmitCommand(
                    self.component_exec_name,
                    SystemEvents.DISPLAY.value,
                    dict(action=self.action_name),
                ).mount(self.processor)
            )
        return commands

    def __repr__(self):
        return "CallDisplayCommand({}, {})".format(
            self.component_exec_name, self.action_name
        )


class CallListenerCommand(Command):
    def __init__(
        self,
        component_exec_name: str,
        listener_name: str,
        event: "Event",
    ):
        super().__init__(component_exec_name)
        self.listener_name = listener_name
        self.event = event

    def execute(self):
        component = self.processor.components[self.component_exec_name]
        cconfig = component._config
        if self.listener_name not in cconfig.component_listeners:
            raise JembeError(
                "Listener {}.{} does not exist".format(
                    cconfig.full_name, self.listener_name
                )
            )

        # execute listener
        listener_result = getattr(component, self.listener_name)(self.event)
        component._jembe_has_action_or_listener_executed = True
        if listener_result is None or (
            isinstance(listener_result, bool) and listener_result == True
        ):
            # after executing listener that returns True or None
            # component should be rendered by executing display
            self.processor.add_command(
                CallDisplayCommand(component.exec_name, force=listener_result == True),
                end=True,
            )
        elif isinstance(listener_result, bool) or listener_result == False:
            # Do nothing
            pass
        elif isinstance(listener_result, Response):
            # TODO return response and
            # stop fruther processiong of requst with jembe processor
            # TODO jembe.js should be able to handle direct reponses
            # TODO make the same for action
            # raise NotImplementedError()
            return listener_result
        else:
            raise JembeError(
                "Invalid listener result type: {}.{} {}".format(
                    cconfig.full_name,
                    self.listener_name,
                    listener_result,
                )
            )

    def __repr__(self):
        return "CallListener({}, {}, {})".format(
            self.component_exec_name, self.listener_name, self.event
        )


class EmitCommand(Command):
    # TODO match listner extract as method
    def __init__(
        self,
        component_exec_name: str,
        event_name: str,
        params: dict,
        to: Optional[Union[str, Sequence[str]]] = None,
    ):
        super().__init__(component_exec_name)
        self.event_name = event_name
        self.params = params
        self._to: Tuple[str, ...] = ()
        if to is not None:
            self.to(to)

        self.primary_execution = True

        self.event: "Event"

    def to(self, to: Optional[Union[str, Sequence[str]]]) -> "EmitCommand":
        """
        emit_event_to is glob like string for finding component.

        if emit_event_to is:

            - None -> match to every initialised component
            - /compoent1.key1/compoent2.key2    -> compoenent with complete exec_name
            - ./component                       -> emit to direct child named "component" without key
            - ./component.*                     -> emit to direct child named "component" with any key
            - ./component.key                   -> emit to direct child named "component with key equals "key"
            - ./**/component[.[*|<key>]]        -> emit to child at any level
            - ..                                -> emit to parent
            - ../component[.[*|<key>]]          -> emit to sibling
            - /**/.                             -> emit to parent at any level
            - /**/component[.[*|<key>]]/**/.    -> emit to parent at any level named
            - //                                -> emit to my root page TODO
            - etc.
        """
        if to is None:
            pass
        elif isinstance(to, str):
            self._to = (to,)
        else:
            self._to = tuple(to)
        return self

    def execute(self):
        """
        finds all components mathcing self.to that have registred listeners
        whose source is matched to self.component.exec_name and calls matching
        listeners
        """
        self.event = Event(
            source_exec_name=self.component_exec_name,
            source=self.processor.components.get(self.component_exec_name, None),
            event_name=self.event_name,
            to=self._to,
            params=self.params,
        )
        execute_over: List[Tuple["jembe.Component", str]] = []
        for exec_name, component in self.processor.components.items():
            for (
                listener_method_name,
                listener,
            ) in component._config.component_listeners.items():
                if self._is_match(
                    source_exec_name=self.component_exec_name,
                    event_name=self.event_name,
                    source_to=self._to,
                    destination_exec_name=component.exec_name,
                    destination_listener=listener,
                ):
                    execute_over.append((component, listener_method_name))

        # order components from top to bottom if message is send only to parents
        # this is required for proper handling of exceptions
        if self._to == "/**/.":
            execute_over.sort(key=lambda t: t[0]._config.hiearchy_level, reverse=True)
        for comp, listener_method_name in execute_over:
            self.processor.add_command(
                CallListenerCommand(comp.exec_name, listener_method_name, self.event)
            )

        if self.primary_execution:
            self.processor._emited_event_commands.append(self)

    @classmethod
    def _glob_match_exec_name(
        cls, pattern_exec_name, pattern, component_exec_name
    ) -> bool:
        """
        Check if glob pattern set by component with pattern_exec_name match
        with compoment_exec_name

        if pattern is:

            - None -> match to every initialised component
            - /compoent1.key1/compoent2.key2    -> compoenent with complete exec_name
            - *                               -> direct children with or without key
            - *.*                             -> direct children with any key
            - **/*                            -> all children with or without key
            - **/*.*                          -> all children with any key
            - component                       -> match to direct child named "component" without key
            - component.*                     -> match to direct child named "component" with any key
            - component.key                   -> match to direct child named "component with key equals "key"
            - **/component[.[*|<key>]]        -> match to child at any level
            - ..                                -> match to parent
            - ../component[.[*|<key>]]          -> match to sibling
            - /**/.                             -> match to parent at any level
            - /**/component[.[*|<key>]]/**/.    -> match to parent at any level named
            - //                                -> process event from root page
            - .                                 -> component that emited event
            - etc.
        """

        if pattern is None:
            # match any component
            return True

        if "/**/." in pattern:
            # reverse match for parent
            if pattern_exec_name == component_exec_name:
                return False
            if not pattern_exec_name.startswith(component_exec_name):
                return False
            if not pattern.startswith("/**/") or pattern == "/**/.":
                return True
            else:
                re_pattern = (
                    "(/{p}/)|(/{p}$)".format(
                        p=re.escape(
                            # removes /**/ from begining and /**/. from end
                            pattern[4:-5]
                        )
                    )
                    .replace(
                        "/\\*\\*/",
                        "((/.*/)|(/))",  # ** replace wit regex to match compoente path
                    )
                    .replace("\\*\\.\\*", "[^./]+\\.[^./]+")
                    .replace(
                        "\\.\\*", "\\.[^./]+"
                    )  # replace with regex to match any key
                    .replace("\\*", "[^/]*")  # replace to match compoent name
                )
                return re.search(re_pattern, pattern_exec_name) is not None
        else:
            # regular path match
            if pattern.startswith("/"):
                pass
            elif pattern == ".":
                pattern = pattern_exec_name
            elif not (pattern.startswith(".") or pattern.startswith("/")):
                pattern = "{}/{}".format(pattern_exec_name, pattern)
            elif pattern.startswith("./"):
                pattern = "{}{}".format(pattern_exec_name, pattern.lstrip("."))
            elif pattern.startswith(".."):
                pattern_begins = pattern_exec_name.lstrip("/").split("/")
                while pattern.startswith(".."):
                    pattern = pattern.lstrip(".").lstrip("/")
                    try:
                        pattern_begins = pattern_begins[:-1]
                    except KeyError:
                        # no parrent so no match
                        return False
                pattern = "/".join(["", *pattern_begins, pattern]).rstrip("/")

            re_pattern = (
                re.escape(pattern)
                .replace(
                    "/\\*\\*/",
                    "((/.*/)|(/))",  # ** replace wit regex to match component path
                )
                .replace("\\*\\.\\*", "[^./]+\\.[^./]+")
                .replace("\\.\\*", "\\.[^./]+")  # replace with regex to match any key
                .replace("\\*", "[^/]*")  # replace to match compoent name
            )
            return re.fullmatch(re_pattern, component_exec_name) is not None

    def __repr__(self):
        return "Emit({}, {}, {})".format(
            self.component_exec_name, self.event_name, self._to
        )

    def reemit_over(
        self, reemit_component_exec_name: str
    ) -> List["CallListenerCommand"]:
        reemit_component = self.processor.components[reemit_component_exec_name]
        commands = list()
        for (
            listener_method_name,
            listener,
        ) in reemit_component._config.component_listeners.items():
            if self._is_match(
                source_exec_name=self.component_exec_name,
                event_name=self.event_name,
                source_to=self._to,
                destination_exec_name=reemit_component_exec_name,
                destination_listener=listener,
            ):
                commands.append(
                    CallListenerCommand(
                        reemit_component.exec_name,
                        listener_method_name,
                        self.event,
                    )
                )
        return commands

    @classmethod
    def _is_match(
        cls,
        source_exec_name: str,
        event_name: str,
        source_to: Tuple[str, ...],
        destination_exec_name: str,
        destination_listener: "ComponentListener",
    ) -> bool:
        def __is_match(
            _source_to: str,
            listener_event_name: Optional[str],
            listener_source: Optional[str],
        ) -> bool:
            return (
                (
                    # if listener_event_name is None only catch user event but not
                    # jembe system events that starts with underscore (_)
                    (listener_event_name is None and not event_name.startswith("_"))
                    or listener_event_name == event_name
                )
                and cls._glob_match_exec_name(
                    # match emit.to with compoenent name
                    source_exec_name,
                    _source_to,
                    destination_exec_name,
                )
                and cls._glob_match_exec_name(
                    # match for compoennt filter on listener end
                    destination_exec_name,
                    listener_source,
                    source_exec_name,
                )
            )

        sources_to: tuple = tuple((None,)) if not source_to else source_to
        listener_events: tuple = (
            tuple((None,))
            if not destination_listener.event_names
            else destination_listener.event_names
        )
        listener_sources: tuple = (
            tuple((None,))
            if not destination_listener.sources
            else destination_listener.sources
        )
        for s_to in sources_to:
            for dl_event in listener_events:
                for dl_source in listener_sources:
                    if __is_match(s_to, dl_event, dl_source):
                        return True
        return False


class InitialiseCommand(Command):
    def __init__(
        self,
        component_exec_name: str,
        init_params: dict,
        merge_existing_params: bool = True,
        exists_on_client: bool = False,
    ):
        super().__init__(component_exec_name)
        self._init_params = {
            k: (v if not isinstance(v, Undefined) else None)
            for k, v in init_params.items()
        }
        self.merge_existing_params = merge_existing_params
        self.exists_on_client = exists_on_client
        self.displayed_components: List[str] = []

        self.initialised_component: Optional["jembe.Component"] = None
        self._cconfig: "jembe.ComponentConfig"

    def mount(self, processor: "Processor") -> "InitialiseCommand":
        self._cconfig = processor.jembe.get_component_config(self.component_exec_name)
        return super().mount(processor)

    @cached_property
    def existing_component(self) -> Optional["jembe.Component"]:
        if self.component_exec_name in self.processor.components_marked_for_removal:
            return None
        return self.processor.components.get(self.component_exec_name, None)

    @cached_property
    def init_params(self) -> dict:
        existing_params = (
            {
                k: v
                for k, v in self.existing_component.state.items()
                if k
                not in self.existing_component.state._injected_params_names  # TODO: Test this and figure out why it's needed?
            }
            if self.merge_existing_params and self.existing_component
            else dict()
        )
        init_params = (
            {}
            if self.merge_existing_params
            else {**self._cconfig.component_class._jembe_state_param_default_values}
        )
        init_params.update(
            {**existing_params, **self._init_params, **self._inject_into_params}
        )
        return init_params

    @cached_property
    def _inject_into_params(self) -> Dict[str, Any]:
        parent_cconfig = self._cconfig.parent
        if parent_cconfig:
            parent_component = self.processor.components[
                parent_exec_name(self.component_exec_name)
            ]
            injected_params = (
                parent_component.inject_into(self._cconfig)
                if parent_component._jembe_inject_into_overriden
                else dict()
            )
            if parent_cconfig._inject_into_components:
                injected_params.update(
                    parent_cconfig._inject_into_components(
                        parent_component,
                        self._cconfig,
                    )
                )
            # clean up injected params
            if injected_params:
                injected_params = {
                    key: value
                    for key, value in injected_params.items()
                    if key in self._cconfig.component_class._jembe_init_param_names
                }
            return injected_params
        else:
            return dict()

    def _must_do_init(self, is_accessible_run: bool):
        if self.existing_component:
            # new_params = (
            #     {}
            #     if self.merge_existing_params
            #     else {**self._cconfig.component_class._jembe_state_param_default_values}
            # )
            # new_params.update(
            #     {
            #         **self._init_params,
            #         **self._inject_into_params,
            #         **self.existing_component.inject(),
            #     }
            # )
            # if state params are same continue
            # else raise jembeerror until find better solution
            has_new_params = False
            # for k, v in new_params.items():
            for k, v in self.init_params.items():
                if (
                    k in self.existing_component.state
                    and v != self.existing_component.state[k]
                ):
                    has_new_params = True
                    break

            if self.existing_component._jembe_has_action_or_listener_executed:
                if has_new_params:
                    if is_accessible_run:
                        return True
                    else:
                        raise JembeError(
                            (
                                "Cant reinitialise component {} with new parametes after "
                                "action or listener is executed by this component.\n"
                                "Existing state: {}\n"
                                "New state: {}"
                            ).format(
                                self.component_exec_name,
                                self.existing_component.state,
                                self.init_params,
                            )
                        )
                else:
                    return False
            else:
                return has_new_params
        return True

    def execute(self, is_accessible_run=False):
        # create new component if component with identical exec_name does not exist
        # or component with identical exec_name has not action or listener executed
        if self._must_do_init(is_accessible_run):
            component = self._cconfig.component_class._jembe_init_(
                self._cconfig,
                self.component_exec_name,
                list(self._inject_into_params.keys()),
                self.merge_existing_params,
                self.existing_component.state if self.existing_component else None,
                **self.init_params
            )

            # check if action passes access control
            if not component.ac_check():
                raise ComponentConfig.DEFAULT_AC_EXCEPTION()

            self.initialised_component = component
            if not is_accessible_run:
                self.processor.components[component.exec_name] = component

                if self.exists_on_client:
                    self.processor.renderers[component.exec_name] = ComponentRender(
                        False,
                        component.state.tojsondict(component, True),
                        component._jembe_disabled_actions.copy(),
                        component.state._injected_params_names,
                        component.url,
                        component._config.changes_url,
                        None,
                        self.displayed_components,
                    )

                if self.processor._emited_event_commands:
                    # we need to reemit commands to this component
                    for emited_command in self.processor._emited_event_commands:
                        reemit_commands = emited_command.reemit_over(
                            self.component_exec_name
                        )
                        for remit_cmd in reemit_commands:
                            self.processor.add_command(remit_cmd, end=True)
        else:
            self.initialised_component = self.processor.components[
                self.component_exec_name
            ]

    def get_before_emit_commands(self) -> Sequence["EmitCommand"]:
        return ()

    def get_after_emit_commands(self) -> Sequence["EmitCommand"]:
        return ()

    def __repr__(self):
        try:
            return "Init({}, {})".format(self.component_exec_name, self.init_params)
        except:
            return "Init({}, {})".format(self.component_exec_name, self._init_params)


class ComponentRender(NamedTuple):
    """Represents html rendered by component with additional parametars"""

    fresh: bool
    state_jsondict: Dict[str, Any]
    disabled_actions: List[str]
    injected_params: List[str]
    url: Optional[str]
    changes_url: bool
    html: Optional[str]
    displayed_components: List[str]


class CommandsQue:
    def __init__(self, jembe: "jembe.Jembe") -> None:
        self.commands: Deque["Command"] = deque()
        self.deferred_commands: Deque["Command"] = deque()

        self.jembe = jembe

    def add_command(self, command: "Command", end=False) -> None:
        """
        Adds new command into staginig commands que.
        if end is True command is added at the end of que (last to execute)

        Staging que exist in order to allow commands generated by while
        execution one command to be internally ordered by appending at
        the beginig or the end of staging que

        It also addes all before and after commands in staging que
        """
        if not command.is_mounted:
            raise JembeError("Cann't add unmounted command to commands que")

        def _do_add_command(que, command, end):
            """Adds command with before and after command to stack"""
            before_emit_commands = command.get_before_emit_commands()

            if end:
                # adds to the end of que .. last to execute
                for before_cmd in before_emit_commands:
                    que.appendleft(before_cmd)

                que.appendleft(command)

            else:
                # adds at the begingin of que .. first to execute
                que.append(command)

                for before_cmd in reversed(before_emit_commands):
                    que.append(before_cmd)

        # Check if command should be deffered
        is_deferred_command = False
        if isinstance(command, CallActionCommand):
            try:
                cconfig = self.jembe.components_configs[command.component_full_name]
            except AttributeError:
                raise JembeError(
                    "Component {} does not exist".format(command.component_full_name)
                )
            try:
                caction = cconfig.component_actions[command.action_name]
            except AttributeError:
                raise JembeError(
                    "Action {} is not defined for component {}".format(
                        command.action_name, command.component_full_name
                    )
                )
            is_deferred_command = caction.deferred

        if is_deferred_command:
            _do_add_command(self.deferred_commands, command, end)
        else:
            _do_add_command(self.commands, command, end)

    def move_commands_to(self, que: Deque["Command"]) -> None:
        """Moves commands from staging que to execution que"""
        while self.deferred_commands:
            que.append(self.deferred_commands.popleft())
        while self.commands:
            que.append(self.commands.popleft())

    def clear(self) -> None:
        self.commands.clear()
        self.deferred_commands.clear()

    def __repr__(self) -> str:
        return "CommandsQue({}, defered={})".format(
            self.commands, self.deferred_commands
        )


class Processor:
    """
    1. Will use deapest component.url from all components on the page as window.location
    2. When any component action or listener returns False default action
       (display) will not be called
    """

    def __init__(
        self, _jembe: "jembe.Jembe", component_full_name: str, request: "Request"
    ):
        g.jmb_processor = self

        self.jembe = _jembe
        self.request = request

        self.components: Dict[str, "jembe.Component"] = dict()
        self._commands: Deque["Command"] = deque()
        self._processing_command: Optional["Command"] = None
        # already emited and processed event commands
        # that will be executed over newelly initialised component
        self._emited_event_commands: Deque["EmitCommand"] = deque()
        # commands created while executing some command and that needs to be
        # added to commands at the end of command execution
        self._staging_commands = CommandsQue(self.jembe)
        # component renderers is dict[exec_name] = (componentState, url, rendered_str)
        self.renderers: Dict[str, "ComponentRender"] = dict()
        # list of execnames marked for removal by jembe client js without redisplaying parent
        self.components_marked_for_removal: List[str] = []
        # component that raised exception on initialise
        # any subsequent command on this component should be ignored
        self._raised_exception_on_initialise: Dict[str, dict] = dict()
        # direct response if component display returns it
        self._response: Optional["Response"] = None

        # List of init commands without coresponding call command that
        # need to be checked for redisplay depending of other redisplayed compoents
        # on page (when processing x-jembe request)
        self._hanging_init_commands_execnames: List[str] = []
        self.__create_commands(component_full_name)
        self._staging_commands.move_commands_to(self._commands)

    def add_command(self, command: "Command", end=False) -> None:
        self._staging_commands.add_command(
            command if command.is_mounted else command.mount(self), end
        )

    def _load_init_params(self, exec_name: str, init_params: dict) -> dict:
        component_config = self.jembe.get_component_config(exec_name)
        component_class = component_config.component_class

        load_params = dict()
        for param_name, param_value in init_params.items():
            load_params[param_name] = component_class.load_init_param(
                component_config, param_name, param_value
            )
        return load_params

    def _x_jembe_command_factory(self, command_data: dict) -> "Command":
        """Process data received by json and build commands

        is_component - mark that command_data is received in components section of x-jembe request
                    describing current compoents and their state displayed to the user
        """

        if "execName" in command_data:
            return InitialiseCommand(
                command_data["execName"],
                self._load_init_params(command_data["execName"], command_data["state"]),
                exists_on_client=True,
            )
        elif command_data["type"] == "init":
            return InitialiseCommand(
                command_data["componentExecName"],
                self._load_init_params(
                    command_data["componentExecName"], command_data["initParams"]
                ),
                merge_existing_params=command_data["mergeExistingParams"],
            )
        elif (
            command_data["type"] == "call"
            and command_data["actionName"] == ComponentConfig.DEFAULT_DISPLAY_ACTION
        ):
            return CallDisplayCommand(command_data["componentExecName"])
        elif (
            command_data["type"] == "call"
            and command_data["actionName"] != ComponentConfig.DEFAULT_DISPLAY_ACTION
        ):
            return CallActionCommand(
                command_data["componentExecName"],
                command_data["actionName"],
                command_data["args"],
                command_data["kwargs"],
            )
        elif command_data["type"] == "emit":
            return EmitCommand(
                command_data["componentExecName"],
                command_data["eventName"],
                command_data["params"],
                command_data["to"],
            )
        raise NotImplementedError()

    def __create_commands(self, component_full_name: str):
        if self._is_x_jembe_request:
            try:
                data = json.loads(
                    self.request.data.decode("utf-8"),
                )
            except JSONDecodeError:
                data = dict()
        if self._is_x_jembe_request and data:
            # x-jembe ajax request
            # init components from data["components"]
            to_be_initialised = [
                component_data["execName"] for component_data in data["components"]
            ]
            for component_data in data["components"]:
                initcmd = cast(
                    "InitialiseCommand", self._x_jembe_command_factory(component_data)
                )
                initcmd.displayed_components = [
                    en
                    for en in to_be_initialised
                    if is_direct_child_name(initcmd.component_exec_name, en)
                ]
                self.add_command(initcmd, end=True)
            # init components from url_path if thay doesnot exist in data["compoenents"]
            self.__create_commands_from_url_path(component_full_name, to_be_initialised)

            # init commands
            updated_init_commands_execnames: List[str] = []
            call_commands_execnames: List[str] = []
            for command_data in data["commands"]:
                command = self._x_jembe_command_factory(command_data)
                self.add_command(command, end=True)
                if isinstance(command, InitialiseCommand):
                    updated_init_commands_execnames.append(command.component_exec_name)
                elif isinstance(command, CallActionCommand):
                    call_commands_execnames.append(command.component_exec_name)
            # find all init commands without coresponding call or display command
            self._hanging_init_commands_execnames = [
                c
                for c in updated_init_commands_execnames
                if c not in call_commands_execnames
            ]
        else:
            # regular http/s GET request and when invalid x-jembe request data
            exec_names = self.__create_commands_from_url_path(
                component_full_name, list()
            )

            self.add_command(
                CallDisplayCommand(exec_names[-1]),
                end=True,
            )
            for exec_name in exec_names[:-1]:
                self.add_command(
                    CallDisplayCommand(exec_name),
                    end=True,
                )

    def __create_commands_from_url_path(
        self, component_full_name: str, to_be_initialised: List[str]
    ) -> List[str]:
        """
        inits components from request url_path
        if component with same exec_name is not already initialised

        returns exec names from root to component_full_name
        """
        from .component import Component

        exec_names = []

        c_configs: List["ComponentConfig"] = list()
        # get all components configs from root to requested component
        for cname in component_full_name.strip("/").split("/"):
            c_configs.append(
                self.jembe.components_configs[
                    "/".join((c_configs[-1].full_name if c_configs else "", cname))
                ]
            )
        # initialise all components from root to requested component
        # >> parent component exec name, makes easy to build next component exec_name
        parent_exec_name = ""
        for cconfig in c_configs:
            if cconfig.name is None:
                raise JembeError()

            request_view_args = (
                self.request.view_args if self.request.view_args is not None else dict()
            )
            key = request_view_args[cconfig._key_url_param.identifier].lstrip(".")
            exec_name = Component._build_exec_name(cconfig.name, key, parent_exec_name)
            parent_exec_name = exec_name
            if exec_name not in to_be_initialised:
                init_params = {
                    up.name: request_view_args[up.identifier]
                    for up in cconfig._url_params
                }
                if (
                    not self._is_x_jembe_request
                    and exec_name_to_full_name(exec_name) == component_full_name
                ):
                    # do mapping for get query url params
                    for (
                        url_param_name,
                        state_param_name,
                    ) in cconfig.url_query_params.items():
                        value = self.request.args.get(url_param_name, None)
                        if value is not None:
                            try:
                                init_params[state_param_name] = load_param(
                                    cconfig.component_class._jembe_init_signature.parameters[
                                        state_param_name
                                    ].annotation,
                                    unquote_plus(self.request.args[url_param_name]),
                                )
                            except ValueError as ve:
                                current_app.logger.warning(
                                    "{}: {} {}".format(
                                        ve,
                                        exec_name,
                                        cconfig.component_class._jembe_init_signature.parameters[
                                            state_param_name
                                        ],
                                    )
                                )
                                pass

                self.add_command(
                    self._x_jembe_command_factory(
                        dict(
                            type="init",
                            componentExecName=exec_name,
                            initParams=init_params,
                            mergeExistingParams=True,
                        )
                    ),
                    end=True,
                )
            exec_names.append(exec_name)

        if exec_name not in self.request.path:
            # dont allow that /mainaaaa open /main.aaaa
            raise NotFound()
        return exec_names

    def process_request(self) -> "Processor":
        try:
            response = self._execute_commands()

            if response is not None:
                self._delete_tmp_uploads()
                self._response = response
                return self

            # for all freshly rendered component who does not have parent renderers
            # eather at client (send via x-jembe.components) nor just created by
            # server call display command
            needs_render_exec_names = set(
                chain.from_iterable(
                    accumulate(
                        map(lambda x: "/" + x, exec_name.strip("/").split("/")), add
                    )
                    for exec_name, cr in self.renderers.items()
                    if cr.fresh == True
                )
            )
            missing_render_exec_names = needs_render_exec_names - set(
                self.renderers.keys()
            )
            for exec_name in sorted(
                missing_render_exec_names,
                key=lambda exec_name: self.components[exec_name]._config.hiearchy_level,
            ):
                self.add_command(CallDisplayCommand(exec_name))
            self._staging_commands.move_commands_to(self._commands)

            self._execute_commands()
            # for all components that have change state params but not have been redisplayed
            # and all components with RedisplayFlag.WHEN_ON_PAGE:
            # check will thay still be presented/visible on page if so execute display command
            # for that components
            for hanging_init_execname in self._hanging_init_commands_execnames + [
                k
                for k, v in self.components.items()
                if RedisplayFlag.WHEN_ON_PAGE in v._config.redisplay
            ]:
                if hanging_init_execname not in [
                    execname for execname, v in self.renderers.items() if v.fresh
                ]:
                    # if hanging init is displayed on page it should be rendered
                    en_split = hanging_init_execname.split("/")
                    execname_tree = []
                    for i in range(2, len(en_split) + 1):
                        execname_tree.append("/".join(en_split[:i]))
                    # check if root component will be redisplayed
                    root_fresh_keys = [
                        k
                        for k, v in self.renderers.items()
                        if v.fresh and len(k.split("/")) == 2
                    ]
                    displayed_execnames = (
                        root_fresh_keys
                        if len(root_fresh_keys) > 0
                        else [execname_tree[0]]
                    )
                    execute_display = True
                    for ename in execname_tree:
                        if ename not in displayed_execnames:
                            execute_display = False
                            break
                        displayed_execnames = self.renderers[ename].displayed_components
                    if execute_display:
                        self.add_command(CallDisplayCommand(hanging_init_execname))
                        self._staging_commands.move_commands_to(self._commands)
                        self._execute_commands()
                    elif (
                        hanging_init_execname in self.components
                        and RedisplayFlag.WHEN_ON_PAGE
                        in self.components[hanging_init_execname]._config.redisplay
                        and hanging_init_execname
                        not in self.components_marked_for_removal
                    ):
                        self.components_marked_for_removal.append(hanging_init_execname)
            return self

        finally:
            self._delete_tmp_uploads()

    def _execute_commands(self) -> Optional["Response"]:
        """executes all commands from self._commands que"""
        while self._commands:
            response = self._execute_command(self._commands.pop())
            if response is not None:
                return response
        return None

    def _execute_command(self, command: "Command") -> Optional["Response"]:
        # command is over component that raised exception on initialise and
        # it is not new initialise command over that commponent, so we skip its execution
        if command.component_exec_name in self._raised_exception_on_initialise:
            if (
                isinstance(command, InitialiseCommand)
                and command.init_params
                == self._raised_exception_on_initialise[command.component_exec_name]
            ):
                return None
            if (
                not isinstance(command, InitialiseCommand)
                and command.component_exec_name not in self.components
            ):
                return None

        try:
            self._processing_command = command
            response = command.execute()
            self._processing_command = None
            self._staging_commands.move_commands_to(self._commands)
            if response is not None:
                return response
        except JembeError as jmberror:
            # JembeError are exceptions raised by jembe
            # and thay indicate bad usage of framework and
            # thay should not be raised in production
            raise jmberror
        except Exception as exc:
            self._handle_exception_in_command(command, exc)
            return None
        else:
            # If execution of command is successfull then
            # add after commands into command que
            for after_cmd in reversed(command.get_after_emit_commands()):
                self.add_command(after_cmd)
            self._staging_commands.move_commands_to(self._commands)
            return None

    def execute_initialise_command_successfully(
        self, command: "InitialiseCommand", *additional_components: "jembe.Component"
    ) -> Tuple[bool, Optional["jembe.Component"]]:
        """
        Directly out of commands que execute initialise command
        in order to check will it raise Exception

        no additional after or before commands will be executed
        """
        # check if component with requested full_name exist or if
        # initialisation with same init_params already failed

        if (
            exec_name_to_full_name(command.component_exec_name)
            not in self.jembe.components_configs
        ):
            return (False, None)
        command = command if command.is_mounted else command.mount(self)
        if command.component_exec_name in self._raised_exception_on_initialise and (
            command.init_params
            == self._raised_exception_on_initialise[command.component_exec_name]
        ):
            return (False, None)

        # execute initialise command without running before or after commands
        backup_current_staging_commands = self._staging_commands
        self._staging_commands = CommandsQue(self.jembe)
        backup_current_components = self.components.copy()
        # add additional components into components
        for acomp in additional_components:
            self.components[acomp.exec_name] = acomp

        try:
            command.execute(is_accessible_run=True)
        except JembeError as jmb_error:
            # JembeError are exceptions raised by jembe
            # and thay indicate bad usage of framework
            self.components = backup_current_components
            raise jmb_error
        except Exception as exc:
            self._raised_exception_on_initialise[
                command.component_exec_name
            ] = deepcopy(command.init_params)
            # initalise command is not run properly

            # restore _staging_commands
            self._staging_commands = backup_current_staging_commands
            self.components = backup_current_components
            if current_app.debug or current_app.testing:
                current_app.logger.warning(
                    "Exception when initialising component out of proccessing que {}: {}".format(
                        command, exc.__repr__()
                    )
                )
                if not isinstance(
                    exc, (AccessDenied, NotFound, Unauthorized, Forbidden)
                ):
                    import traceback

                    traceback.print_exc()
            return (False, None)
        self._staging_commands = backup_current_staging_commands
        self.components = backup_current_components
        return (True, command.initialised_component)

    def _handle_exception_in_command(self, command: "Command", exc: "Exception"):
        """
        Exception that ocure while executing command (call action, display, listener invocation etc)
        are handled in following way.

        Event('_exception', source=command.component_exec_name, params=dict(exception=exc, handled=False)) is created
        and emited to all its parent components from top to bottom.
        This is only time when ordering for receiving messages is guaranted, so that root components can know if
        exception is already handled by its children.

        Parent component can use regular listener to catch event do whatever thay want. @listener('_exception')
        If any of parrent component mark event as handled event.params["handled"] = True exception will not be raised again.
        IMPORTANT: When parent component mark exception as handled all its subcomponents and events will be removed from processor
        so that parent component can rerender itself and/or its subcomponents from the blank state

        If there is not listener for _exception event in parent compoent or exception event is not handled. Processor
        will rerise exception and it will be handled like regular flask exception.
        """
        # Clear all staging commands because command has raised exception
        self._staging_commands.clear()

        # Create emit command to emit exception to component parents
        emit_command = (
            EmitCommand(
                command.component_exec_name,
                SystemEvents.EXCEPTION.value,
                dict(
                    exception=exc,
                    handled=False,
                    in_action=command.action_name
                    if isinstance(command, CallActionCommand)
                    else None,
                ),
            )
            .to((".", "/**/."))
            .mount(self)
        )
        emit_command.execute()
        # self._staging_commands is fill with calllistener commands
        # move it to exception_commands que that will be used
        # to process exception
        exception_commands: Deque["Command"] = deque()
        self._staging_commands.move_commands_to(exception_commands)
        # executes commands from exception_commands que and
        # all new generated commands will be saved in now empty
        # staging_commands que
        while exception_commands:
            # simple loop no catchin exception when handlin exception
            # commands from exception commands dont have after command
            cmd = exception_commands.pop()
            # if exception is raised during handling exception ... dont try to catch it :)
            cmd.execute()
            if emit_command.event.params["handled"]:
                if isinstance(command, InitialiseCommand):
                    # save initialises command that raised exception with params
                    # so we can skip executing other commants on this component
                    # (save to skip execution because we handled exception)
                    self._raised_exception_on_initialise[
                        command.component_exec_name
                    ] = deepcopy(command.init_params)
                    # remove all child components of command.component_exec_name
                    self.components = {
                        exec_name: comp
                        for exec_name, comp in self.components.items()
                        if not is_child_name(cmd.component_exec_name, exec_name)
                    }
                    # remove all commands to and from removed components
                    self._commands = deque(
                        (
                            c
                            for c in self._commands
                            if not is_child_name(
                                cmd.component_exec_name, c.component_exec_name
                            )
                        )
                    )
                exception_commands = deque()
                self._staging_commands.move_commands_to(self._commands)
                return
            else:
                self._staging_commands.move_commands_to(exception_commands)

        # if exception is not handled by parent components
        # rerise exception
        if current_app.debug or current_app.testing or not isinstance(exc, NotFound):
            current_app.logger.error(
                "Unhandled exception in {}: {}".format(command, exc)
            )
            import traceback

            traceback.print_exc()
        raise emit_command.event.params["exception"]

    def build_response(self) -> "Response":
        if self._response:
            return self._response
        # compose full page
        if self._is_x_jembe_request:
            ajax_responses = []
            for (
                exec_name,
                (
                    fresh,
                    state_jsondict,
                    disabled_actions,
                    injected_params_names,
                    url,
                    changes_url,
                    html,
                    components,
                ),
            ) in self.renderers.items():
                if fresh and exec_name not in self.components_marked_for_removal:
                    ajax_responses.append(
                        dict(
                            execName=exec_name,
                            state={
                                k: v
                                for k, v in state_jsondict.items()
                                if k not in injected_params_names
                            },
                            dom=html,
                            url=url,
                            changesUrl=changes_url,
                            actions={
                                action_name: action_name not in disabled_actions
                                for action_name in self.jembe.get_component_config(
                                    exec_name
                                ).component_actions.keys()
                                if action_name != ComponentConfig.DEFAULT_DISPLAY_ACTION
                            },
                        )
                    )
            if self.components_marked_for_removal:
                ajax_responses.append(
                    dict(
                        globals=True,
                        removeComponents=self.components_marked_for_removal,
                    )
                )
            return jsonify(ajax_responses)
        else:
            # for page with components build united response
            c_etrees = {
                exec_name: self._lxml_add_dom_attrs(
                    html,
                    exec_name,
                    {
                        k: v
                        for k, v in state_jsondict.items()
                        if k not in injected_params_names
                    },
                    url,
                    changes_url,
                    disabled_actions,
                )
                for exec_name, (
                    fresh,
                    state_jsondict,
                    disabled_actions,
                    injected_params_names,
                    url,
                    changes_url,
                    html,
                    components,
                ) in self.renderers.items()
                if fresh
                and state_jsondict is not None
                and url is not None
                and html is not None
                and exec_name not in self.components_marked_for_removal
            }
            unused_exec_names = sorted(
                c_etrees.keys(),
                key=lambda exec_name: self.components[exec_name]._config.hiearchy_level,
            )
            response_etree = None
            can_find_placeholder = True
            while unused_exec_names and can_find_placeholder:
                can_find_placeholder = False
                if response_etree is None:
                    response_etree = c_etrees[unused_exec_names.pop(0)]
                # compose response including all components not just page
                # find all placeholders in response_tree and replace them with
                # appropriate etrees
                placeholders = []
                permanent_placeholders = []
                for placeholder in response_etree.xpath(
                    ".//template[@jmb-placeholder or @jmb-placeholder-permanent]"
                ):
                    if "jmb_placeholder" in placeholder.attrib:
                        placeholders.append(placeholder)
                    else:
                        permanent_placeholders.append(placeholder)

                for placeholder in placeholders + permanent_placeholders:
                    try:
                        permanent = False
                        exec_name = placeholder.attrib["jmb-placeholder"]
                    except KeyError:
                        permanent = True
                        exec_name = placeholder.attrib["jmb-placeholder-permanent"]

                    if exec_name in unused_exec_names:
                        can_find_placeholder = True
                        unused_exec_names.pop(unused_exec_names.index(exec_name))
                        try:
                            c_etree = c_etrees[exec_name]
                            placeholder.addnext(c_etree)
                        except KeyError:
                            # exec_name referenced by this placeholder does not exist
                            # so we will just remove placeholder
                            # This situation can ocure when handling exceptions
                            # of child components but not changing display html
                            # so we can assume that developer just want to ignore
                            # exception and dont display component that coused exception
                            pass
                        finally:
                            if not permanent:
                                placeholder.getparent().remove(placeholder)

            # Remove empty placeholder if thay are left in response
            # because above logic will not find all empty placeholders
            if response_etree is not None:
                for placeholder in response_etree.xpath(
                    ".//template[@jmb-placeholder]"
                ):
                    placeholder.getparent().remove(placeholder)
            return etree.tostring(response_etree, method="html")

    def _lxml_add_dom_attrs(
        self,
        html: str,
        exec_name: str,
        state_jsondict: Dict[str, Any],
        url: str,
        changes_url: bool,
        disabled_actions: List[str],
    ):  # -> "lxml.html.HtmlElement":
        """
        Adds dom attrs to html.
        If html has one root tag attrs are added to that tag othervise
        html is souranded with div
        """

        def set_jmb_attrs(elem):
            elem.set("jmb-name", exec_name)
            elem.set(
                "jmb-data",
                json.dumps(
                    dict(
                        actions={
                            action_name: action_name not in disabled_actions
                            for action_name in self.jembe.get_component_config(
                                exec_name
                            ).component_actions.keys()
                            if action_name != ComponentConfig.DEFAULT_DISPLAY_ACTION
                        },
                        changesUrl=changes_url,
                        state=state_jsondict,
                        url=url,
                    ),
                    separators=(",", ":"),
                    sort_keys=True,
                    # default=json_default
                ),
            )

        if not html.strip():
            html = "<div></div>"
        root = etree.HTML(html)
        if is_page_exec_name(exec_name):
            # exec_name is fom page component, component with no parent
            doc = root.getroottree()
            set_jmb_attrs(root)
            return doc
        else:
            doc = root[0]
            if (
                len(root[0]) == 1
                and "jmb-placeholder" not in root[0][0].attrib
                and "jmb-name" not in root[0][0].attrib
            ):
                set_jmb_attrs(root[0][0])
            else:
                # add div tag at root[0][0] and move all root[0][0..len] to new tag
                div = Element("div")
                set_jmb_attrs(div)
                children = root[0].getchildren()
                root[0][0].addprevious(div)
                for child in children:
                    div.append(child)
            return doc[0]

    @cached_property
    def _is_x_jembe_request(self) -> bool:
        return bool(
            self.jembe.X_JEMBE in self.request.headers
            and self.request.headers.get(self.jembe.X_JEMBE) != "upload"
        )

    @cached_property
    def _is_x_jembe_upload_followup_request(self) -> bool:
        return (
            self._is_x_jembe_request
            and self.jembe.X_RELATED_UPLOAD in self.request.headers
        )

    @cached_property
    def is_x_jembe_upload_request(self) -> bool:
        return bool(self.request.headers.get(self.jembe.X_JEMBE, None) == "upload")

    def _delete_tmp_uploads(self):
        from jembe.app import get_temp_storage

        if self._is_x_jembe_upload_followup_request:
            upload_request_id = self.request.headers.get(self.jembe.X_RELATED_UPLOAD)
            ts = get_temp_storage()
            try:
                ts.rmtree(ts.path_join("uploads", upload_request_id))
            except:
                # TODO log error
                pass
            try:
                ts.rmdir("uploads")
            except:
                pass

    def remove_component(self, exec_name: str, only_children: bool = False):
        def match_exec_name(match: str) -> bool:
            if match.startswith("{}/".format(exec_name)):
                return True
            if only_children == False and match == exec_name:
                return True
            return False

        self._commands = deque(
            (
                c
                for c in self._commands
                if not match_exec_name(c.component_exec_name)
                or isinstance(c, EmitCommand)
            )
        )
        self.renderers = {
            en: r for en, r in self.renderers.items() if not match_exec_name(en)
        }
        self.components = {
            en: c for en, c in self.components.items() if not match_exec_name(en)
        }



from typing import (
    TYPE_CHECKING,
    Iterable,
    Optional,
    Union,
    Dict,
    Any,
    List,
    Tuple,
    Type,
)
import re
from urllib.parse import quote_plus
from functools import cached_property
from copy import deepcopy, copy
from abc import ABCMeta
from inspect import Parameter, signature, getmembers, Signature

from flask.json import dumps
from .exceptions import JembeError, NotFound, ComponentPreviousStateUnavaiableError
from flask import render_template, render_template_string, current_app
from markupsafe import Markup
from .component_config import ComponentConfig
from .app import get_processor
from .processor import (
    CallActionCommand,
    CallDisplayCommand,
    InitialiseCommand,
    EmitCommand,
)
from .common import (
    exec_name_to_full_name,
    load_param,
    dump_param,
)

if TYPE_CHECKING:
    import jembe


class ComponentState(dict):
    """
    Two instance of the same component, component with same full_name, in the same
    state should behawe identically.

    State of the component is defined by state params, __init__ params whose name
    doesn't start with _ (underscore).

    Appon initialising component_state params can't be added or deleted from
    that state but values of the existing params can change
    """

    def __init__(self, *args, **kwargs):
        self._injected_params_names: List[str] = []
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        if not key in self.keys():
            raise JembeError(f"Can't add new param '{key}' to component state {self}")
        return super().__setitem__(key, value)

    def __delitem__(self, key):
        raise JembeError("Can't delete param from component state")

    def __getattr__(self, name):
        if name in self.keys():
            return self[name]
        return super().__getattribute__(name)

    def __setattr__(self, name, value):
        if name == "_injected_params_names":
            super().__setattr__(name, value)
        else:
            if name not in self.keys():
                raise JembeError("Can't set arbitrary attribute to component state")
            self[name] = value
        # return super().__setattr__(name, value)

    def __eq__(self, value):
        return super().__eq__(value)

    def deepcopy(self):
        c = ComponentState(**self)
        for key, value in self.items():
            c[key] = deepcopy(value)
        c._injected_params_names = deepcopy(self._injected_params_names)
        return c

    def tojsondict(
        self,
        component_class: Union[Type["jembe.Component"], "jembe.Component"],
        full=False,
    ):
        return {
            k: copy(component_class.dump_init_param(k, v))
            for k, v in self.items()
            if full == True or k not in self._injected_params_names
        }


class ComponentReference:
    """
    Notes:
    usage for link:
    lamda self: self.component('name')[.jrl...]
    lambda self: self.component('/main').component('name')[.jrl|url|is_accessible]
    usage for displaying component in template:
    {{component('name')}}
    {{component('name').component('name')}} -- shuld raise exception
    usage in template to display link
    {{component('name').is_accessible}}..{{component().jrl}}...
    Add comand 'init' with same usage as component but without
    using state params of exisiting component if its on page ??


    factory class method allow use of {{component('../name')}} or {{component('/a/b/c')}}
    """

    @classmethod
    def factory(
        cls,
        caller_exec_name: Optional[str],
        name: str,
        kwargs: dict,
        merge_existing_params: bool = True,
    ) -> "ComponentReference":
        name_split = name.split("/")
        cr: Optional["ComponentReference"] = None
        if (name_split[0] == "" and len(name_split) > 2) or (  # /a/b
            name_split[0] != "" and len(name_split) > 1
        ):  # a/b ../b
            if name_split[0] == "":
                name_split = name_split[1:]
                name_split[0] = f"/{name_split[0]}"
            for pname in name_split[:-1]:
                if cr is None:
                    cr = cls(
                        caller_exec_name,
                        pname,
                        {},
                        # merge_existing_params,
                    )
                else:
                    cr = cr.component(pname)
            if cr is not None:
                cr = (
                    cr.component(name_split[-1], **kwargs)
                    if merge_existing_params
                    else cr.component_reset(name_split[-1], **kwargs)
                )
            else:
                cr = cls(
                    caller_exec_name, name_split[-1], kwargs, merge_existing_params
                )
        elif "/" not in name and "." not in name and "()" in name:
            # compoent("cancel()").is_accessible
            cr = cls(caller_exec_name, ".", dict(), merge_existing_params).call(
                name.replace("()", ""), **kwargs
            )
        else:
            cr = cls(caller_exec_name, name, kwargs, merge_existing_params)
        return cr

    def __init__(
        self,
        caller_exec_name: Optional[str],
        name: str,
        kwargs: dict,
        merge_existing_params: bool = True,
    ):
        """
        - name can be name of subcomponent or if it starts with '/' name of root component
        """
        self._key = ""

        if name == "":
            raise JembeError("Component name can't be empty string")

        if name == "." and caller_exec_name is None:
            raise JembeError(
                "ComponentReference can't be called  to reference self without caller_exec_name specified"
            )

        self.caller_exec_name = caller_exec_name
        self.name = name
        self.action = ComponentConfig.DEFAULT_DISPLAY_ACTION
        self.action_args: Tuple[Any, ...] = ()
        self.action_kwargs: dict = {}
        self.state_kwargs = {k: v for k, v in kwargs.items() if not k.startswith("_")}
        self.kwargs = {k: v for k, v in kwargs.items()}
        self.merge_existing_params = merge_existing_params

        self._component_initialise_done = False
        self._component_instance: Optional["Component"] = None
        self._is_accessible: Optional[bool] = None

        if self.caller_exec_name is None and not self.name.startswith("/"):
            raise JembeError(
                "ComponentReference can't be called with relative path without caller_exec_name specified"
            )

        if (".." in self.name and self.name != "..") or (
            "/" in self.name
            and not self.name.startswith("/")
            # and len(self.name.split("/")) == 2
        ):
            raise JembeError(
                "Component reference only supports rendering and accessing direct childs, root page or parent component"
            )

        self.root_renderer = self
        self.active_renderer = self
        self.parent_reference = self
        self._aditional_components: List["Component"] = []
        self.base_jrl = "$jmb"

    def _component(
        self, merge_existing_params, jmb_exec_name: str, **kwargs
    ) -> "jembe.ComponentReference":
        cr = ComponentReference(
            self.exec_name, jmb_exec_name, kwargs, merge_existing_params
        )
        cr.root_renderer = self.root_renderer
        cr.root_renderer.active_renderer = cr
        cr.parent_reference = self
        cr._aditional_components = self._aditional_components.copy()
        if (
            self.component_instance is not None
            and self.component_instance.exec_name not in self.processor.components
        ):
            cr._aditional_components.append(self.component_instance)
        cr.base_jrl = self.jrl
        return cr

    def component(self, jmb_exec_name: str, **kwargs) -> "jembe.ComponentReference":
        if self.action != ComponentConfig.DEFAULT_DISPLAY_ACTION:
            raise JembeError(
                f"<{str(self)}> Cant reference child component when action '{self.action}' is called"
            )
        return self._component(True, jmb_exec_name, **kwargs)

    def component_reset(
        self, jmb_exec_name: str, **kwargs
    ) -> "jembe.ComponentReference":
        return self._component(False, jmb_exec_name, **kwargs)

    def _init_component(self):
        if self._component_initialise_done:
            return
        self._component_initialise_done = True

        if not (self.name == "." and self.root_renderer == self):
            initialise_command = InitialiseCommand(
                self.exec_name, self.kwargs, self.merge_existing_params
            )
            (
                self._is_accessible,
                self._component_instance,
            ) = self.processor.execute_initialise_command_successfully(
                initialise_command, *self._aditional_components
            )

    @property
    def component_instance(self) -> Optional["jembe.Component"]:
        if self.name == "." and self.root_renderer == self:
            return self.processor.components[self.exec_name]
        else:
            self._init_component()
            return self._component_instance

    @property
    def is_accessible(self) -> bool:
        # TODO add param ignore_incoplete_params = True so that exception trown during initialise
        # becouse not all required init parameters are suplied are treated as not accessible (
        # catch exception and return False in this case)
        if self.action != ComponentConfig.DEFAULT_DISPLAY_ACTION:
            ci = self.component_instance
            if ci is None:
                current_app.logger.warning(
                    f"Component '{self.root_renderer.caller_exec_name}' can't acuire referenced component '{self.exec_name}' to check if action '{self.action}' is accessible"
                )
                return True

            return (
                self.action in ci._config.component_actions
                and self.action not in ci._jembe_disabled_actions
            )
        else:
            if self.name == ".":
                return True
            self._init_component()
            return (
                self._is_accessible == True
                and self._component_instance is not None
                and self._component_instance.ac_check()
            )

    @cached_property
    def url(self) -> str:
        if not self.is_accessible:
            raise NotFound()
        if self._aditional_components:
            backup_components = self.processor.components.copy()
            for acomp in self._aditional_components:
                self.processor.components[acomp.exec_name] = acomp
            url = self.component_instance.url  # type: ignore
            self.processor.components = backup_components
            return url
        else:
            return self.component_instance.url  # type: ignore

    @cached_property
    def jrl(self) -> str:
        if not self.is_accessible:
            raise NotFound()
        jrl = (
            "component{reset}('{name}'{kwargs}{key})".format(
                reset="_reset" if not self.merge_existing_params else "",
                name=self.name,
                key=f",key='{self.key}'" if self._key else "",
                kwargs=",{}".format(
                    re.sub(
                        '(?<!\\\\)"',
                        "'",
                        dumps(self.state_kwargs, separators=(",", ":")),
                    )
                )
                if self.state_kwargs
                else "",
            )
            if self.name != "."
            else ""
        )
        jrl += (
            # ".call('{name}',{{{kwargs}}},[{args}])".format(
            ".call('{name}'{kwargs})".format(
                name=self.action,
                # args=",".join((_prep_v(v) for v in self.action_args)),
                kwargs=",{}".format(
                    re.sub(
                        '(?<!\\\\)"',
                        "'",
                        dumps(self.action_kwargs, separators=(",", ":")),
                    )
                )
                if self.action_kwargs
                else "",
            )
            if self.action != ComponentConfig.DEFAULT_DISPLAY_ACTION
            else ".display()"
        )
        if jrl.startswith("."):
            jrl = jrl[1:]
        base_jrl = (
            self.base_jrl[:-10]
            if self.base_jrl.endswith(".display()")
            else self.base_jrl
        )
        return Markup(f"{base_jrl}.{jrl}")

    @cached_property
    def exec_name(self) -> str:
        if self.name == ".":
            return self.caller_exec_name  # type:ignore
        elif self.name.startswith("/"):
            return Component._build_exec_name(self.name.split("/")[1], self._key)
        elif self.caller_exec_name is not None and self.name == "..":
            caller_exec_name_split = self.caller_exec_name.split("/")
            return Component._build_exec_name(
                caller_exec_name_split[-2],
                self._key,
                "/".join(caller_exec_name_split[:-2]),
            )
        elif self.caller_exec_name is not None:
            return Component._build_exec_name(
                self.name, self._key, self.caller_exec_name
            )
        else:
            raise JembeError(
                "ComponentReference can't be called with relative path without caller_exec_name specified"
            )

    @cached_property
    def processor(self) -> "jembe.Processor":
        return get_processor()

    def execute(self):
        """Calls all necessary init, display and call components to execute this component reference
        inside current jembe processor.
        """
        if self.exec_name in self.processor.components_marked_for_removal:
            raise JembeError(
                f"Cant display component '{self.exec_name}' marked for removal in parent component template!"
            )
        if self.parent_reference != self:
            self.parent_reference.execute()

        self.processor.add_command(
            InitialiseCommand(self.exec_name, self.kwargs, self.merge_existing_params),
            end=True,
        )
        # self._component_instance = None  # stop using component from is_accessible check

        # call action command is put in que to be executed latter
        # if this command raises exception parent should chach it and call display
        # with appropriate template
        if self.action == ComponentConfig.DEFAULT_DISPLAY_ACTION:
            self.processor.add_command(
                CallDisplayCommand(
                    self.exec_name,
                    not self.merge_existing_params,
                    displayed_by_exec_name=self.caller_exec_name,
                ),
                end=True,
            )
        else:
            self.processor.add_command(
                CallActionCommand(
                    self.exec_name,
                    self.action,
                    self.action_args,
                    self.action_kwargs,
                ),
                end=True,
            )

    def __call__(self) -> str:
        """
        Do render

        Component initialisation must be executed in place in order to know
        will it raise exception (like NotFound, Forbidden, Unauthorized etc.)
        so that we can decide how to render template
        """
        self.execute()
        return Markup(f'<template jmb-placeholder="{self.exec_name}"></template>')

    def key(self, key: str) -> "jembe.ComponentReference":
        self._key = key
        return self

    def call(self, action: str, *args, **kwargs) -> "jembe.ComponentReference":
        self.action = action
        self.action_args = args
        self.action_kwargs = kwargs
        return self

    def __html__(self):
        return self.__call__()

    def if_accessible(self) -> str:
        if self.is_accessible:
            return self.__html__()
        return ""


def component(
    jmb_exec_name: str, jmb_reset: bool = True, **kwargs
) -> "ComponentReference":
    """Creates component renderer that can be used to obtain any component url, jrl or check if it is accessible"""
    return ComponentReference.factory(None, jmb_exec_name, kwargs, not jmb_reset)


def componentInitDecorator(init_method):
    def decoratedInit(self, *args, **kwargs):
        """
        Inject params from inject method and saves states params in component state
        """
        if not hasattr(self, "state"):
            # only top most class in hieararchy sets componentState and injects from self.inject()

            # Inject params from inject method
            params_to_inject = self.inject()
            self._jembe_injected_params_names.extend(params_to_inject.keys())

            for name, value in params_to_inject.items():
                if current_app.debug and name in kwargs:
                    current_app.logger.warning(
                        f"Injecting already set param {name} in component {self._config.full_name}"
                    )
                kwargs[name] = value

            # Saves states params in component state
            self.state = ComponentState(
                **{
                    spn: kwargs.get(
                        spn, self._jembe_init_signature.parameters[spn].default
                    )
                    for spn in self._jembe_state_param_names
                }
            )
            self.state._injected_params_names = self._jembe_injected_params_names
        try:
            init_method(self, *args, **kwargs)
        except Exception as e:
            current_app.logger.warning(
                f"{self.__class__.__name__}: {e};  args={args}; kwargs={kwargs};"
            )
            raise e

    return decoratedInit


class ComponentMeta(ABCMeta):
    def __new__(cls, name, bases, attrs, **kwargs):
        # decorate __init__ to create self.state varible from init params
        if "__init__" in attrs:
            # attach original init signature to bu used by component config
            # to define url params
            init_signature = signature(attrs["__init__"])
            attrs["_jembe_init_signature"] = init_signature
            attrs["_jembe_init_param_names"] = tuple(
                p.name for p in init_signature.parameters.values() if p.name != "self"
            )
            attrs["_jembe_state_param_names"] = tuple(
                p.name
                for p in init_signature.parameters.values()
                if p.name != "self" and not p.name.startswith("_")
            )
            attrs["_jembe_state_param_default_values"] = {
                p.name: p.default
                for p in init_signature.parameters.values()
                if p.default != Parameter.empty
                and p.name != "self"
                and not p.name.startswith("_")
            }
            attrs["__init__"] = componentInitDecorator(attrs["__init__"])

        # check if inject_into is overriden
        inject_into_overriden = False
        if "inject_into" in attrs:
            inject_into_overriden = True
        else:
            for b in bases:
                if b != Component and getattr(b, "_jembe_inject_into_overriden", False):
                    inject_into_overriden = True
                    break
        attrs["_jembe_inject_into_overriden"] = inject_into_overriden

        new_class = super().__new__(cls, name, bases, dict(attrs), **kwargs)
        return new_class


class Component(metaclass=ComponentMeta):
    """Component is the primary building block for creating Web Applications using Jembe Framework.

    Component is responsible:
        - for displaying part of the web page, and
        - handling all user interaction with that part of the page.

    Component is made from two parts:
        - Python class that holds and executes application logic, and
        - associated Jinja template that renderes, component in browser;


    Component can:
        - be comined with other components in nested hirearchy to form web application;
        - execute "actions" when end user interacts with component in browser;
        - dispatch events to other components in application;
        - listen for events dispatched to them;

    .. note::

        Component behavior can be configured with Component.Config class that is initialised only once when the application starts.

    Component has **State variables**. State variables defines component state. Two instance of same component are
    equal if they are having equal state variables.


    - State variable are passed to component on its initialisation;
    - Same component with same values of state variables are considered equal (thay should display exacly same HTML in browser)
    - When value of any state variable is changed component will redisplay itself;
    - State variables are defined as arguments to __init__ method.
    - State varibales must have type annotation, and type annotation cannot be in double quotes.
    - State variables are directly accessible to JavaScript in browser;
    - State variables can be changed by JavaScript in browser cosing component to redisplay itself.

    .. code-block:: python

        @jmb.page('hello')
        class HelloWorld(Component):
            def __init__(self, name:str = "World"):
                ''' name is state variable '''
                pass

    .. code-block:: html

        <html><body>
            <h1>Hello {{name}}</h1>
            <input type="text" value="{{name}}" jmb-on:keydown.debounce="name = $self.value">

            <script src="{{ url_for('jembe.static', filename='js/jembe.js') }}"></script>
        </body></html>

    .. figure:: /img/hello_world.gif
        :alt: Hello World

    **Component __init__ arguments**

    Component init argumets are used to define:

    - **state variables** and,
    - performance variables.

    **State variables** are __init__ arguments that does not start with '_' (underscore).
    State variables defines the state of the component. Two instance of the same component in the same state are
    considered equal and thay should behaive in the exacly same way.

    **Performance variables** are __init__ arguments whose name starts with underscore '_'.
    Performance variables are used to avoid unnecesary calculations or call to database, when parent component
    has already calculated or aquiried value needed by component.

    State variables without default values will become **URL arguments**. Value of Url arguments
    must be provided as part of URL (for example: https://your_site/project/project/1).
    State variables without default values, can be one of following types: int, str, URI.

    .. code-block:: python

        @jmb.page('demo')
        class CDemo(Component):
            def __init__(self, id: int, find:Optional[str] = None, _record:Optional[dict] = None):
                '''
                    - 'id' is State variable of type int.
                    - 'id' is also URL variable because it does not have default value.
                      When browser access:

                      -  https://your_site/demo/1 id will have value of 1
                      -  https://your_site/demo/2 id will have value of 2, and so on

                    - 'find' is state variable of type optional str with default value of None.
                      This variable can't be set via url like 'id'.
                    - '_record' is performance variable it is not state variable.
                '''
                if record['id'] = id:
                    # if _record is valid use it
                    self._record = record

            @property
            def record(self):
                try:
                    return self._record
                except:
                    self._record = db.session.query(Record).get(self.state.id)
                    return self._record


    Args:
        metaclass (_type_, optional): _description_. Defaults to ComponentMeta.

    Raises:
        JembeError: When something goes wrong with component initialisation
    """

    @classmethod
    def _jembe_init_(
        cls,
        _config: ComponentConfig,
        _component_exec_name: str,
        _jembe_injected_params_names: List[str],
        _jembe_merged_existing_params: bool,
        _jembe_existing_component_state: Optional["ComponentState"],
        **init_params,
    ):
        """
        Instance creation by explicitly calling __new__ and __init__
        because _config should be avaiable in __init__
        """
        component: "Component" = object.__new__(cls)
        component._jembe_component_initialising = True
        component._config = _config  # type: ignore
        component._jembe_injected_params_names = _jembe_injected_params_names
        component._jembe_merged_existing_params = _jembe_merged_existing_params
        component._jembe_existing_component_state = _jembe_existing_component_state
        component._jembe_disabled_actions = []
        component.exec_name = _component_exec_name
        component.__init__(**init_params)  # type: ignore
        # init is called after __init__ so it can access all state variables regardles if __init__ is overwriten
        component.init()
        component._jembe_component_initialising = None
        component._jembe_existing_component_state = None
        return component

    _jembe_init_signature: "Signature"
    _jembe_init_param_names: Tuple[str, ...]
    _jembe_state_param_names: Tuple[str, ...]
    _jembe_state_param_default_values: Dict[str, Any]
    _jembe_injected_params_names: List[str]
    _jembe_config_init_params: Dict[str, Any]
    _jembe_merged_existing_params: bool
    _jembe_existing_component_state: Optional["ComponentState"]
    _jembe_inject_into_overriden: bool
    _jembe_component_initialising: Optional[bool]
    _config: "Config"

    state: "ComponentState"
    _jembe_disabled_actions: List[str]

    class Config(ComponentConfig):
        """Compononent config defines behavior of all instances of component, with what is
        known at build time, like: url_path, subcomponents, name etc.

        Component can access his config instance using ``self._config``.

        Args:
            template (Optional[Union[str, Iterable[str]]], optional): Path to default template for displaying component. If not provided template name will be same as ``self.full_name`` with '.html' extension added. Defaults to None.
            components (Optional[Dict[str, jembe.ComponentRef]], optional): Subcomponent definitions. Defaults to None.
            inject_into_components (Optional[ Callable[[jembe.Component, jembe.ComponentConfig], dict] ], optional): Callable to inject __init__ paramas into subcomponents. Defaults to None.
            changes_url (bool, optional): Does this componet changes location URL when displayed on page. Defaults to True.
            url_query_params (Optional[Dict[str, str]], optional): Mapping from GET Query params to state variables allowing state variables to be set with GET Query params (?var1=value&var2=value).dict(<name of get queryparam> = <name of state variable>) . Defaults to None.
        """

        pass

    def __init__(self):
        try:
            self.__key: str = self.__key
        except:
            self.__key = ""
        self.__exec_name: str
        self.__has_action_or_listener_executed: bool = False

    def init(self):
        """
        Normalised __init__ without params definition.

        init is called after __init__ so it can access all state variables
        regardles if __init__ is overwriten.

        Executed after __init__, usefull for adding logic that should be
        executed on init but without need to rewrite/list init params.
        """
        # raise NotFound when componet is not accessible
        if not self.ac_check():
            raise self._config.DEFAULT_AC_EXCEPTION()

    @property
    def previous_state(self) -> Optional["ComponentState"]:
        """Returns previous state of the component when component state is changed in browser using javascript.

        Accesible only in ``__init__`` and ``init`` methods.
        """
        if not self._jembe_component_initialising:
            raise ComponentPreviousStateUnavaiableError(
                "Component previous state is only avaiable during "
                "component intialisation in __init__ and init methods."
            )
        return self._jembe_existing_component_state

    @property
    def key(self) -> str:
        """Returns component key.

        ``key`` can be set by parent component, allowing us to have multiple
        instance of same component in different state. Parent component
        must make sure that there is not two component with same key.

        Usefull when displaying list of records with cards.

        If we try to display multiple components in diferent state without setting key,
        only last component will be displayed overwriting all previous ones.

        Returns:
            str: component key
        """
        return self.__key

    @key.setter
    def key(self, key: str):
        self.__key = key
        # self.__dict__.pop("key", None) # delete cached property
        # TODO update __exec_name
        # self.__update_exec_name()

    @property
    def exec_name(self) -> str:
        """Uniquly identifies instance of the Component in application.

        It is created from component ``full_name`` adding all ``keys`` where necessary
        (/component1/component2.key2/component3/component4.key4).

        Returns:
            str: Component Execution Name
        """
        return self.__exec_name

    @exec_name.setter
    def exec_name(self, exec_name: str):
        # verify exec_name with _config.full_name
        if exec_name_to_full_name(exec_name) != self._config.full_name:
            raise JembeError(
                f"Invalid exec_name {exec_name} for {self._config.full_name}"
            )

        self.__exec_name = exec_name

        # set , __key
        name_key = exec_name.strip("/").split("/")[-1].split(".")
        if len(name_key) == 2:
            self.__key = name_key[1]

    @property
    def _jembe_has_action_or_listener_executed(self) -> bool:
        return self.__has_action_or_listener_executed

    @_jembe_has_action_or_listener_executed.setter
    def _jembe_has_action_or_listener_executed(self, value: bool):
        if value != True:
            raise JembeError(
                "Cannot reset action or listener execution status on component"
            )
        self.__has_action_or_listener_executed = True

    def set_parent_exec_name(self, parent_exec_name: str):
        self.__exec_name = f"{parent_exec_name}/{self._config.name}.{self.key}"
        # TODO verify parent_exec_name with _config.full_name
        # TODO set __exec_name

    @classmethod
    def _build_exec_name(
        cls, name: str, key: str = "", parent_exec_name: str = ""
    ) -> str:
        """Build component exec name"""
        local_exec_name = name if not key else f"{name}.{key}"
        return "/".join((parent_exec_name, local_exec_name))

    @property
    def url(self) -> str:
        """Returns url of this component.

        Url is build using url_path of parent components and url_path of this component
        """
        url = self._config.build_url(self.exec_name, self)
        url_get_params = []
        for url_param_name, state_param_name in self._config.url_query_params.items():
            if self.state.get(state_param_name, None) is not None:
                url_get_params.append(
                    f"{url_param_name}={quote_plus(str(self.state[state_param_name]))}"
                )
        if url_get_params:
            url = f"{url}?{'&'.join(url_get_params)}"
        return url

    @classmethod
    def dump_init_param(cls, name: str, value: Any) -> Any:
        """
        Encode state param for sending to client (dumped param will be  transformed to json).

        dumped value (return value) should be build using only following types:

        - dict, list, tuple, str, int, float, init- & float-deriveted Enums, True, False, None
        so that it can be jsonified

        """
        if name in cls._jembe_init_signature.parameters:
            try:
                param_hint = cls._jembe_init_signature.parameters[name]
                if param_hint.annotation == Parameter.empty:
                    raise ValueError("Parameter without annotation")
                return dump_param(param_hint.annotation, value)
            except Exception as e:
                if current_app.debug or current_app.testing:
                    raise JembeError(
                        f"State param {name} of {cls.__module__}.{cls.__name__} with hint {cls._jembe_init_signature.parameters[name]} is not supported for json dump/load "
                        f"nor custom encoding/decoding logic is defined in dump_init_param/load_init_param.\n ({e})\n\n "
                    )
        return value

    @classmethod
    def load_init_param(cls, config: "jembe.ComponentConfig", name: str, value: Any) -> Any:
        """
        load and Decode init/state param received via json call to be uset to initialise in __init__.
        param_value is decoded from json received from client.

        Default implemntation suports:
            - dict, list, tuple, str, int, float, init- & float-deriveted Enums, True, False, None, via
              default json load.
            - if any other type hint (or no hint) is set for state param and we are running in
              debug mode exception will be raised (No exception will be raised in production
              because hint checking can be expensive)
        """
        if name in cls._jembe_init_signature.parameters:
            try:
                param_hint = cls._jembe_init_signature.parameters[name]
                if param_hint.annotation == Parameter.empty:
                    raise ValueError("Parameter without annotation")
                return load_param(param_hint.annotation, value)
            except ValueError as e:
                if current_app.debug or current_app.testing:
                    raise JembeError(
                        f"State param {name} of {cls.__module__}.{cls.__name__} with hint {cls._jembe_init_signature.parameters[name]} is not supported for json dump/load "
                        f"nor custom encoding/decoding logic is defined in dump_init_param/load_init_param. \n\n({e}) "
                    )
        return value

    def inject(self) -> Dict[str, Any]:
        """
        inject params are used to inject cross functional params into component.
        This params usually defines enviroment in which component is exected and
        are not related nor handled by parent compoenents or url params.

        Typical examples of inject params are user_id, userObject of current user usually stored
        in session. Current user language or similar that is stored in session/cookie
        or passed as header param to every request.

        Component will inject this paramas into its __init__ method, it will ignore
        existing values of params set by default, manually or via parent component.
        If for some reason injected params are explicitly set they will be ignored in
        production but in development warrning will be displayed.

        inject can set any type of init param except for url param (params without default value).
        if some param is injected it value will not be send to client nor acepted from x-jembe
        http request.
        """
        return dict()

    def inject_into(self, cconfig: "jembe.ComponentConfig") -> Dict[str, Any]:
        """
        inject_into params are used to inject  params into child component.
        This params usually defines values required by child component that are
        generated or optained by parent component.

        inject_into is called when child component is initialised, which can
        happend right after parent comonent is intialised or during execution
        of display on parent component.

        Typical examples of inject_into params are:
        - parent_record_id,

        Child component will inject this paramas into its __init__ method, it will ignore
        existing values of params set by default or manualy (in template or via ajax request),
        but will be overrriden if same param is injected by the child component.

        If for some reason injected_into params are explicitly set manually they will be ignored in
        production but in development warrning will be displayed.
        Also if child component does not accept injected_into param this param will be ignored
        and during development info will be displayed.

        inject into can set any type of init param except for url param (params without default value).
        if some param is injected it value will not be send to client nor acepted from x-jembe
        http request and JembeError will be raised.
        """
        #  Not supportting this becouse it will force reinitialise of all childrens if
        #   parent component action changes state regardles if parent injects params with
        #   this method
        return dict()

    def isinjected(self, param_name: str) -> bool:
        return param_name in self._jembe_injected_params_names

    def display(self) -> "jembe.DisplayResponse":
        return self.render_template()

    def render_template(
        self, template: Optional[Union[str, Iterable[str]]] = None, **context
    ) -> str:
        """Renderes jinja2 template into html, adds default context variables

        TODO
        - context param defines additional template context variables
        - if template_name is none render default template

        TODO IF init_params are the same in two renderes of the same component
        (same full_name and same key and same init_params, excluding init params
        which name starts with underscore) then dont rerender component
        """
        context = {
            **self._get_default_template_context(),
            **context,
        }
        template = template if template else self._config.template
        if not isinstance(template, str):
            template = list(template)
        # add reference to itself in context (usefull for passing parameters in macros)
        context["_context"] = context
        return render_template(template, **context)  # type:ignore

    def render_template_string(self, source, **context):
        """
        Renderes jinja2 template string into html, adds default context variables

        - context param defines additional template context variables


        TODO IF init_params are the same in two renderes of the same component
        (same full_name and same key and same init_params, excluding init params
        which name starts with underscore) then dont rerender component
        """
        context = {
            **self._get_default_template_context(),
            **context,
        }
        # add reference to itself in context (usefull for passing parameters in macros)
        context["_context"] = context
        return render_template_string(source, **context)

    def _get_default_template_context(self) -> Dict[str, Any]:
        """
        returns dict with:
            - state params decomposed in context
            - and all instance variables not starting with underscore including (self.state)

        Note: decomposed state params are overiden by instnace variable if
        instance variable with same name as state param exist
        """
        return {
            # adds state params
            **self.state,
            # add class variables not starting with underscore
            **{
                name: value
                for name, value in vars(self.__class__).items()
                if not name.startswith("_")
            },
            # add instance variables not starting with underscore
            **{
                name: value
                for name, value in vars(self).items()
                if not name.startswith("_")
            },
            # add properties not starting with underscore
            # exec_name, key, url and all user defined properties
            **{
                property_name: getattr(self, property_name)
                for property_name, property in getmembers(
                    self.__class__,
                    lambda o: isinstance(o, property) or isinstance(o, cached_property),
                )
                if property_name != "previous_state"
            },
            # command to render subcomponents
            "component": self._jinja2_component,
            "component_reset": self._jinja2_component_reset,
            "placeholder": self._jinja2_placeholder,
            # add helpers
            "_config": self._config,
        }

    def _component_reference(
        self,
        kwargs: Dict[str, Any],
        name: Optional[str] = None,
        merge_existing_params: bool = True,
    ) -> "jembe.ComponentReference":
        if name is None:
            try:
                return self.__prev_sub_component_renderer.active_renderer
            except AttributeError:
                raise JembeError("Previous component renderer is not set")
        else:
            self.__prev_sub_component_renderer: "ComponentReference" = (
                ComponentReference.factory(
                    self.exec_name, name, kwargs, merge_existing_params
                )
            )
            return self.__prev_sub_component_renderer

    def component(
        self, _jembe_component_name: str = ".", **kwargs
    ) -> "jembe.ComponentReference":
        return self._component_reference(kwargs, _jembe_component_name)

    def component_reset(
        self, _jmb_component_name: str = ".", **kwargs
    ) -> "jembe.ComponentReference":
        return self._component_reference(kwargs, _jmb_component_name, False)

    def _jinja2_component(
        self, _jmb_component_name: Optional[str] = None, **kwargs
    ) -> "jembe.ComponentReference":
        return self._component_reference(kwargs, _jmb_component_name)

    def _jinja2_component_reset(
        self, _jmb_component_name: Optional[str] = None, **kwargs
    ) -> "jembe.ComponentReference":
        return self._component_reference(kwargs, _jmb_component_name, False)

    def _jinja2_placeholder(self, _jmb_copmonent_name: str):
        return Markup(
            f'<template jmb-placeholder-permanent="{self.exec_name}/{_jmb_copmonent_name}"></template>'
        )

    def emit(self, _jmb_event_name: str, **params) -> "EmitCommand":
        processor = get_processor()
        emmit_command = EmitCommand(self.exec_name, _jmb_event_name, params)
        processor.add_command(emmit_command)
        return emmit_command

    def display_component(
        self,
        _jmb_component_name: str,
        _jmb_component_key: Optional[str] = None,
        **init_params,
    ):
        """
        Calls initCommand and displayComand of subcomponent without need to
        redisplay this commponent.
        """
        if _jmb_component_name not in self._config.components.keys():
            raise JembeError(
                f"Component '{_jmb_component_name,}' is not valid sub component of '{self._config.full_name}'!"
            )
        cr = self.component(_jmb_component_name, **init_params)
        if _jmb_component_key:
            cr.key(_jmb_component_key)
        cr.execute()

    def remove_component(
        self, _jmb_component_name: str, _jmb_component_key: Optional[str] = None
    ):
        """
        Marks component for removal from user page (will not be displayed to the user)
        without need to redisplay this entire component.

        TODO: Check if component is rendered with {{component()}} and removed, if so
        raise exception.
        """
        if _jmb_component_name not in self._config.components.keys():
            raise JembeError(
                f"Component '{_jmb_component_name}' is not valid sub component of '{self._config.full_name}'!"
            )
        processor = get_processor()
        cexec_name = "{}/{}".format(
            self.exec_name,
            _jmb_component_name
            if _jmb_component_key is None
            else "{}.{}".format(_jmb_component_name, _jmb_component_key),
        )
        self.emit("_remove", removed_exec_name=cexec_name)
        if cexec_name not in processor.components_marked_for_removal:
            processor.components_marked_for_removal.append(cexec_name)

    def get_storage(self, storage_name: Optional[str] = None) -> "jembe.Storage":
        processor = get_processor()
        return processor.jembe.get_storage(storage_name)

    def redirect_to(self, component_ref: "jembe.ComponentReference"):
        """
        Redirects request to another component removing all unecessory
        components from processing.
        """
        processor = get_processor()
        # Remove all unecessary components that should not be needed after
        # redirect by finding first component that is parent to both this compoment
        # and redirected to compomennt (have same exec_name and same state)
        # and remove all existing components and its assciated commands from
        # jembe processors
        root_cr = component_ref.root_renderer
        processor.remove_component(
            root_cr.exec_name,
            self.exec_name.startswith("{root_cr.exec_name}/"),
        )

        # Creates new commands to init and display for redirect components
        # and puts in processenig que
        component_ref()

    def ac_allow(self, *action_names):
        """
        Allow execution of listed actions.
        If no acction is listed allow execution of all actions.
        """
        if len(action_names) == 0:
            self._jembe_disabled_actions = []
        else:
            for a in action_names:
                try:
                    self._jembe_disabled_actions.remove(a)
                except ValueError:
                    pass

    def ac_deny(self, *action_names):
        """
        Deny execution of listed actions.
        If no action is listed or "display" is listed
        whole component is not accesible (__init__ will raise NotFound)
        """
        if len(action_names) == 0:
            self._jembe_disabled_actions = [ComponentConfig.DEFAULT_DISPLAY_ACTION]
            for a in self._config.component_actions.keys():
                self._jembe_disabled_actions.append(a)
        else:
            for a in action_names:
                if a not in self._jembe_disabled_actions and (
                    a in self._config.component_actions.keys()
                    or a == ComponentConfig.DEFAULT_DISPLAY_ACTION
                ):
                    self._jembe_disabled_actions.append(a)

    def ac_check(self, action_name: Optional[str] = None) -> bool:
        """Check if access control rules are satisfied

        Args:
            action_name (Optional[str], optional): Check if action is accessible, when none it will check is whole component is accessible. Defaults to None.

        Returns:
            bool: returns true if action or whole compoennt is accesible
        """
        # Apply access control rules
        if ComponentConfig.DEFAULT_DISPLAY_ACTION in self._jembe_disabled_actions or (
            action_name is not None and action_name in self._jembe_disabled_actions
        ):
            return False
        return True



from functools import wraps
from typing import List, Dict, Any, Optional
from jembe.common import exec_name_to_full_name
from jembe.exceptions import JembeError
from flask import url_for


def page_url(exec_name: str, url_params: Optional[List[Dict[str, Any]]] = None):
    """Returns url of the Component referenced by exec_name (Component Execution Name).

    Url is genarated without initialising components but by bulidin url dynamicaly.

    .. note:: This method uses Url params names instead of state param names.If not renamed with `ComponentConfig(url_query_params={})` names are the same.

    Args:
        exec_name: Component Execution Name which url we want to obtain
        url_params: Optional list of component url params for every component in exec_name.  Defaults to None.

    """
    if not exec_name.startswith("/"):
        exec_name = "/{}".format(exec_name)
    root_page_name = exec_name.split("/")[1].split(".")[0]
    full_name = exec_name_to_full_name(exec_name)
    components_keys = [
        (".{}".format(cn.split(".")[1]) if len(cn.split(".")) == 2 else "")
        for cn in exec_name.split("/")[1:]
    ]
    endpoint_name = "{bp_name}.{full_name}".format(
        bp_name=root_page_name, full_name=full_name
    )
    no_of_components = len(full_name.split("/")) - 1
    url_values = {}
    for i in range(no_of_components):
        ckey_index = "__{}".format(i) if i > 0 else ""
        url_values["component_key{}".format(ckey_index)] = components_keys[i]
    if url_params:
        for (index, params) in enumerate(url_params):
            for key, value in params.items():
                url_key = "{}__{}".format(key, index) if index > 0 else key
                url_values[url_key] = value

    return url_for(endpoint_name, **url_values)


def run_only_once(_method=None, *, for_state: Optional[str] = None):
    """Decorator for component method that executes method only once and caches the result. 


    First time method is called result is cachend, on any subsequent calls method is not executed
    and cached valuted is returned.

    When for_state is provided than cache will be clear when state variable with same
    name is changed.

    .. warning:: 

        This decorator will cache return value regardles of method arguments values. 
        
    This decorator is usefull when we need to perform costly calculation only once
    and don't want and don't need to do it in __init__.

    .. code-block:: python

        @property
        @run_only_once(for_state="id")
        def record(self):
            return db.session.query(ComplexDBView).filter(ComplexDBView.id==id).first()



    Args:
        _method (_type_, optional): _description_. Defaults to None.
        for_state (Optional[str], optional): name of state variable to monitor for change. Defaults to None.

    Raises:
        JembeError: when for_state is not valid state name

    """
    attrname = "_{}_once_result".format(id(_method))

    def decorator(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            for_state_value = None
            if for_state is not None:
                try:
                    for_state_value = self.state[for_state]
                except KeyError:
                    raise JembeError(
                        "{} {}: for_state for decorator run_only_once has invalide state name".format(
                            self._config.full_name, method.__name__
                        )
                    )

            try:
                return_value, saved_for_state_value = getattr(self, attrname)
                if for_state is None or for_state_value == saved_for_state_value:
                    return return_value
            except AttributeError:
                pass
            # execute and save new value
            setattr(self, attrname, (method(self, *args, **kwargs), for_state_value))
            return getattr(self, attrname)[0]

        return wrapper

    if _method is None:
        return decorator
    else:
        return decorator(_method)



from typing import Any, Dict, Optional
from flask import json
from jembe import (
    Component,
    action,
    listener,
    config,
    NotFound,
    Unauthorized,
    UrlPath,
    ComponentConfig
)
from dataclasses import dataclass, field


def test_wiki(jmb, client):
    """bulding simple wiki app for testing server side processing and api"""

    # data definition
    @dataclass
    class User:
        username: str

    @dataclass
    class WikiPage:
        name: str
        title: str
        parent: Optional[str]

        @property
        def path(self):
            return "{}/{}".format(self.parent, self.name) if self.parent else self.name

    # database
    class WikiDb:
        def __init__(self):
            self.pages: Dict[str, WikiPage] = dict(
                root=WikiPage(name="root", title="Root", parent=None)
            )

        def exist(self, page_path: str) -> bool:
            return page_path in self.pages

        def add(self, page: "WikiPage"):
            self.pages[page.path] = page

        def get(self, page_path: str):
            return self.pages[page_path]

        def get_children(self, page_path: str):
            return [page for page in self.pages.values() if page.parent == page_path]

        def get_root_page(self):
            return self.pages["root"]

    wikidb = WikiDb()

    # service to get currently logged user
    # so we dont need to create login components
    session: dict = dict(user=None)

    # forms
    @dataclass
    class WikiEditForm:
        title: str
        error: Optional[str] = field(default=None, init=False)

    @dataclass
    class WikiAddForm:
        name: str
        title: str
        error: Optional[str] = field(default=None, init=False)

    # components
    class View(Component):
        def __init__(self, page_path: UrlPath):
            if not wikidb.exist(self.state.page_path):
                raise NotFound()
            super().__init__()

        def display(self):
            self.page = wikidb.get(self.state.page_path)
            self.children = wikidb.get_children(self.state.page_path)
            self.emit("set_page_title", title=self.page.title)
            return self.render_template_string(
                """<div><h1>{{page.title}}</h1>"""
                """<div>{% for c in children %}<a jmb-on:click="$jmb.set('page_path','{{c.path}}')">{{c.title}}</a><br>{% endfor %}</div></div>"""
            )

    class Edit(Component):
        def __init__(
            self,
            page_path: UrlPath,
            form: Optional[WikiEditForm] = None,
            user: Optional[User] = None,
        ):
            if not wikidb.exist(page_path):
                raise NotFound()

            if user is None or user.username != "admin":
                raise Unauthorized()

            if form is None:
                self.state.form = WikiEditForm(title=self.page.title)

            super().__init__()

        @classmethod
        def load_init_param(
            cls, config: "ComponentConfig", param_name: str, param_value: Any
        ) -> Any:
            if param_name == "form":
                if param_value is None:
                    return None
                return WikiEditForm(title=param_value.get("title", ""))
            else:
                return super().load_init_param(config, param_name, param_value)

        def inject(self):
            return dict(user=session.get("user", None))

        @action
        def save(self):
            # TODO state.form is dict it should be casted to WikiEditForm from jsonobject on initialise
            is_form_valid = bool(self.state.form.title)
            if is_form_valid:
                self.page.title = self.state.form.title
                self.emit("save")
                return False  # don't redisplay
            self.state.form.error = "Title is required"

        def display(self):
            self.emit("set_page_title", title="Edit: {}".format(self.page.title))
            return self.render_template_string(
                "<h1>Edit: {{page.title}}</h1>"
                """{% if form.error %}<div>{{form.error}}</div>{% endif %}"""
                """<label>Title: <input type="text" value="{{form.title}}" jmb-on.change="$jmb.set('form.title', this.value)"></label>"""
                """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
                """<button type="button" jmb-on:click="$jmb.emit('cancel', to='..')">Cancel</button>"""
            )

        @property
        def page(self) -> "WikiPage":
            try:
                return self._page
            except AttributeError:
                self._page: "WikiPage" = wikidb.get(self.state.page_path)
                return self._page

    class Add(Component):
        def __init__(
            self,
            page_path: UrlPath,
            form: Optional[WikiAddForm] = None,
            user: Optional[User] = None,
        ):
            if not wikidb.exist(page_path):
                raise NotFound()

            if user is None or user.username != "admin":
                raise Unauthorized()

            if form is None:
                self.state.form = WikiAddForm(name="", title="")

            super().__init__()

        @classmethod
        def load_init_param(
            cls, config: "ComponentConfig", param_name: str, param_value: Any
        ) -> Any:
            if param_name == "form":
                if param_value is None:
                    return None
                return WikiAddForm(
                    name=param_value.get("name", ""), title=param_value.get("title", "")
                )
            else:
                return super().load_init_param(config, param_name, param_value)

        def inject(self):
            return dict(user=session.get("user", None))

        @action
        def save(self):
            form_is_valid = bool(self.state.form.name) and bool(self.state.form.title)
            if form_is_valid:
                wpage = WikiPage(
                    name=self.state.form.name,
                    title=self.state.form.title,
                    parent=self.state.page_path,
                )
                wikidb.add(wpage)
                self.state.page_path = wpage.path
                self.emit("save")
                return False  # don't redisplay
            self.state.form.error = "Name and title are required"

        def display(self):
            self.page_title = wikidb.get(self.state.page_path).title
            self.emit("set_page_title", title="Add under {}".format(self.page_title))
            return self.render_template_string(
                "<h1>Add under {{page_title}}</h1>"
                """{% if form.error %}<div>{{form.error}}</div>{% endif %}"""
                """<label>Name: <input type="text" value="{{form.name}}" jmb-on.change="$jmb.set('form.name', this.value)"></label>"""
                """<label>Title: <input type="text" value="{{form.title}}" jmb-on.change="$jmb.set('form.title', this.value)"></label>"""
                """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
                """<button type="button" jmb-on:click="$jmb.emit('cancel', to='..')">Cancel</button>"""
            )

    @config(Component.Config(changes_url=False))
    class PageTitle(Component):
        def __init__(self, title: str = "Wiki"):
            super().__init__()

        @listener(event="set_page_title")
        def on_set_page_title(self, event):
            self.state.title = event.title
            # return False

        @action(deferred=True)
        def display(self):
            return self.render_template_string("<title>{{title}}</title>")

    @jmb.page(
        "wiki",
        Component.Config(
            components=dict(page_title=PageTitle, view=View, edit=Edit, add=Add)
        ),
    )
    class Wiki(Component):
        def __init__(self, mode: str = "view"):
            self.goto: Optional[str] = None
            super().__init__()

        @listener(event="_display", source=["view", "edit", "add"])
        def on_display_child(self, event):
            self.state.mode = event.source._config.name
            self.goto = event.source.state.page_path

        @listener(event=["cancel", "save"], source=["edit", "add"])
        def on_cancel_edit_or_add(self, event):
            self.goto = event.source.state.page_path
            self.state.mode = "view"

        def display(self):
            if self.goto is None:
                self.goto = wikidb.get_root_page().path

            return self.render_template_string(
                "<html><head>{{component('page_title')}}</head>"
                "<body>{{component(mode, page_path=goto)}}</body></html>"
            )

    # initial display
    ######################################
    r = client.get("/wiki")
    assert r.status_code == 200
    assert r.data == (
        """<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN" "http://www.w3.org/TR/REC-html40/loose.dtd">"""
        "\n"
        """<html jmb-name="/wiki" jmb-data=\'{"actions":{},"changesUrl":true,"state":{"mode":"view"},"url":"/wiki"}\'>"""
        """<head><title jmb-name="/wiki/page_title" jmb-data=\'{"actions":{},"changesUrl":false,"state":{"title":"Root"},"url":"/wiki/page_title"}\'>Root</title></head>"""
        """<body>"""
        """<div jmb-name="/wiki/view" jmb-data=\'{"actions":{},"changesUrl":true,"state":{"page_path":"root"},"url":"/wiki/view/root"}\'>"""
        """<h1>Root</h1>"""
        """<div></div>"""
        """</div>"""
        """</body></html>"""
    ).encode("utf-8")

    # try edit root before loggedin
    ######################################
    r = client.post(
        "/wiki",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/wiki", state=dict(mode="view")),
                    dict(execName="/wiki/page_title", state=dict(title="Root")),
                    dict(execName="/wiki/view", state=dict(page_path="root")),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/wiki/edit",
                        initParams=dict(page_path="root"),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/wiki/edit",
                        actionName="display",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 401

    # login
    ######################################
    session["user"] = User("admin")

    # display edit
    ######################################
    r = client.post(
        "/wiki",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/wiki", state=dict(mode="view")),
                    dict(execName="/wiki/page_title", state=dict(title="Root")),
                    dict(execName="/wiki/view", state=dict(page_path="root")),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/wiki/edit",
                        initParams=dict(page_path="root"),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/wiki/edit",
                        actionName="display",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    json_response = json.loads(r.data)
    assert r.status_code == 200
    assert len(json_response) == 3
    assert json_response[0]["execName"] == "/wiki"
    assert json_response[0]["state"] == dict(mode="edit")
    assert json_response[0]["dom"] == (
        """<html><head><template jmb-placeholder="/wiki/page_title"></template></head>"""
        """<body><template jmb-placeholder="/wiki/edit"></template></body></html>"""
    )
    assert json_response[1]["execName"] == "/wiki/page_title"
    assert json_response[1]["state"] == dict(title="Edit: Root")
    assert json_response[1]["dom"] == "<title>Edit: Root</title>"
    assert json_response[2]["execName"] == "/wiki/edit"
    assert json_response[2]["state"] == dict(
        form=dict(error=None, title="Root"), page_path="root"
    )
    assert json_response[2]["dom"] == (
        """<h1>Edit: Root</h1>"""
        """<label>Title: <input type="text" value="Root" jmb-on.change="$jmb.set('form.title', this.value)"></label>"""
        """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
        """<button type="button" jmb-on:click="$jmb.emit('cancel', to='..')">Cancel</button>"""
    )

    # change root title - to invalid title
    ######################################
    r = client.post(
        "/wiki",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/wiki", state=dict(mode="edit")),
                    dict(execName="/wiki/page_title", state=dict(title="Edit: Root")),
                    dict(
                        execName="/wiki/edit",
                        state=dict(
                            form=dict(error=None, title="Root"), page_path="root"
                        ),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/wiki/edit",
                        initParams=dict(
                            form=dict(error=None, title=""), page_path="root"
                        ),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/wiki/edit",
                        actionName="save",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 1
    assert json_response[0]["execName"] == "/wiki/edit"
    assert json_response[0]["state"] == dict(
        form=dict(error="Title is required", title=""), page_path="root"
    )
    assert json_response[0]["dom"] == (
        """<h1>Edit: Root</h1>"""
        """<div>Title is required</div>"""
        """<label>Title: <input type="text" value="" jmb-on.change="$jmb.set('form.title', this.value)"></label>"""
        """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
        """<button type="button" jmb-on:click="$jmb.emit('cancel', to='..')">Cancel</button>"""
    )

    # change root title - valid title - goto view
    ######################################
    r = client.post(
        "/wiki",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/wiki", state=dict(mode="edit")),
                    dict(execName="/wiki/page_title", state=dict(title="Edit: Root")),
                    dict(
                        execName="/wiki/edit",
                        state=dict(
                            form=dict(error="Title is required", title=""),
                            page_path="root",
                        ),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/wiki/edit",
                        initParams=dict(
                            form=dict(error="Title is required", title="New Root"),
                            page_path="root",
                        ),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/wiki/edit",
                        actionName="save",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 3
    assert json_response[0]["execName"] == "/wiki"
    assert json_response[0]["state"] == dict(mode="view")
    assert json_response[0]["dom"] == (
        """<html><head><template jmb-placeholder="/wiki/page_title"></template></head>"""
        """<body><template jmb-placeholder="/wiki/view"></template></body></html>"""
    )
    assert json_response[1]["execName"] == "/wiki/page_title"
    assert json_response[1]["state"] == dict(title="New Root")
    assert json_response[1]["dom"] == ("""<title>New Root</title>""")
    assert json_response[2]["execName"] == "/wiki/view"
    assert json_response[2]["state"] == dict(page_path="root")
    assert json_response[2]["dom"] == ("""<div><h1>New Root</h1><div></div></div>""")

    # add page to root - display
    ######################################
    r = client.post(
        "/wiki",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/wiki", state=dict(mode="view")),
                    dict(execName="/wiki/page_title", state=dict(title="New Root")),
                    dict(execName="/wiki/view", state=dict(page_path="root")),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/wiki/add",
                        initParams=dict(page_path="root"),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/wiki/add",
                        actionName="display",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 3
    assert json_response[0]["execName"] == "/wiki"
    assert json_response[0]["state"] == dict(mode="add")
    assert json_response[0]["dom"] == (
        """<html><head><template jmb-placeholder="/wiki/page_title"></template></head>"""
        """<body><template jmb-placeholder="/wiki/add"></template></body></html>"""
    )
    assert json_response[1]["execName"] == "/wiki/page_title"
    assert json_response[1]["state"] == dict(title="Add under New Root")
    assert json_response[1]["dom"] == ("""<title>Add under New Root</title>""")
    assert json_response[2]["execName"] == "/wiki/add"
    assert json_response[2]["state"] == dict(
        page_path="root", form=dict(error=None, name="", title="")
    )
    assert json_response[2]["dom"] == (
        "<h1>Add under New Root</h1>"
        """<label>Name: <input type="text" value="" jmb-on.change="$jmb.set('form.name', this.value)"></label>"""
        """<label>Title: <input type="text" value="" jmb-on.change="$jmb.set('form.title', this.value)"></label>"""
        """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
        """<button type="button" jmb-on:click="$jmb.emit('cancel', to='..')">Cancel</button>"""
    )

    # add page to root - save valid
    ######################################
    r = client.post(
        "/wiki",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/wiki", state=dict(mode="add")),
                    dict(
                        execName="/wiki/page_title",
                        state=dict(title="Add under New Root"),
                    ),
                    dict(
                        execName="/wiki/add",
                        state=dict(
                            page_path="root", form=dict(error=None, name="", title="")
                        ),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/wiki/add",
                        initParams=dict(
                            page_path="root",
                            form=dict(error=None, name="jembe", title="Jembe Page"),
                        ),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/wiki/add",
                        actionName="save",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 3
    assert json_response[0]["execName"] == "/wiki"
    assert json_response[0]["state"] == dict(mode="view")
    assert json_response[0]["dom"] == (
        """<html><head><template jmb-placeholder="/wiki/page_title"></template></head>"""
        """<body><template jmb-placeholder="/wiki/view"></template></body></html>"""
    )
    assert json_response[1]["execName"] == "/wiki/page_title"
    assert json_response[1]["state"] == dict(title="Jembe Page")
    assert json_response[1]["dom"] == ("""<title>Jembe Page</title>""")
    assert json_response[2]["execName"] == "/wiki/view"
    assert json_response[2]["state"] == dict(page_path="root/jembe")
    assert json_response[2]["dom"] == ("""<div><h1>Jembe Page</h1><div></div></div>""")

    # add invalid page to root
    ######################################
    r = client.post(
        "/wiki",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/wiki", state=dict(mode="add")),
                    dict(
                        execName="/wiki/page_title",
                        state=dict(title="Add under New Root"),
                    ),
                    dict(
                        execName="/wiki/add",
                        state=dict(
                            page_path="root", form=dict(error=None, name="", title="")
                        ),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/wiki/add",
                        initParams=dict(
                            page_path="root",
                            form=dict(error=None, name="", title="Jembe Invalid Page"),
                        ),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/wiki/add",
                        actionName="save",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 1
    assert json_response[0]["execName"] == "/wiki/add"
    assert json_response[0]["state"] == dict(
        form=dict(
            error="Name and title are required", name="", title="Jembe Invalid Page"
        ),
        page_path="root",
    )
    assert json_response[0]["dom"] == (
        "<h1>Add under New Root</h1>"
        "<div>Name and title are required</div>"
        """<label>Name: <input type="text" value="" jmb-on.change="$jmb.set('form.name', this.value)"></label>"""
        """<label>Title: <input type="text" value="Jembe Invalid Page" jmb-on.change="$jmb.set('form.title', this.value)"></label>"""
        """<button type="button" jmb-on:click="$jmb.call('save')">Save</button>"""
        """<button type="button" jmb-on:click="$jmb.emit('cancel', to='..')">Cancel</button>"""
    )

    # view root page with link to jembe page
    ######################################
    r = client.get("/wiki")
    assert r.status_code == 200
    assert r.data == (
        """<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN" "http://www.w3.org/TR/REC-html40/loose.dtd">"""
        "\n"
        """<html jmb-name="/wiki" jmb-data=\'{"actions":{},"changesUrl":true,"state":{"mode":"view"},"url":"/wiki"}\'>"""
        """<head><title jmb-name="/wiki/page_title" jmb-data=\'{"actions":{},"changesUrl":false,"state":{"title":"New Root"},"url":"/wiki/page_title"}\'>New Root</title></head>"""
        """<body>"""
        """<div jmb-name="/wiki/view" jmb-data=\'{"actions":{},"changesUrl":true,"state":{"page_path":"root"},"url":"/wiki/view/root"}\'>"""
        """<h1>New Root</h1>"""
        """<div><a jmb-on:click="$jmb.set('page_path',\'root/jembe\')">Jembe Page</a><br></div>"""
        """</div>"""
        """</body></html>"""
    ).encode("utf-8")

    # add page under jembe page
    ######################################
    r = client.post(
        "/wiki",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/wiki", state=dict(mode="add")),
                    dict(
                        execName="/wiki/page_title",
                        state=dict(title="Add under Jembe Page"),
                    ),
                    dict(
                        execName="/wiki/add",
                        state=dict(
                            page_path="root/jembe",
                            form=dict(error=None, name="", title=""),
                        ),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/wiki/add",
                        initParams=dict(
                            page_path="root/jembe",
                            form=dict(error=None, name="suite", title="Suite Page"),
                        ),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/wiki/add",
                        actionName="save",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 3
    assert json_response[0]["execName"] == "/wiki"
    assert json_response[0]["state"] == dict(mode="view")
    assert json_response[0]["dom"] == (
        """<html><head><template jmb-placeholder="/wiki/page_title"></template></head>"""
        """<body><template jmb-placeholder="/wiki/view"></template></body></html>"""
    )
    assert json_response[1]["execName"] == "/wiki/page_title"
    assert json_response[1]["state"] == dict(title="Suite Page")
    assert json_response[1]["dom"] == ("""<title>Suite Page</title>""")
    assert json_response[2]["execName"] == "/wiki/view"
    assert json_response[2]["state"] == dict(page_path="root/jembe/suite")
    assert json_response[2]["dom"] == ("""<div><h1>Suite Page</h1><div></div></div>""")
    # edit page under existing page
    ######################################
    r = client.post(
        "/wiki",
        data=json.dumps(
            dict(
                components=[
                    dict(execName="/wiki", state=dict(mode="edit")),
                    dict(
                        execName="/wiki/page_title",
                        state=dict(title="Edit Suite Page"),
                    ),
                    dict(
                        execName="/wiki/edit",
                        state=dict(
                            page_path="root/jembe/suite",
                            form=dict(error=None, title="Suite Page"),
                        ),
                    ),
                ],
                commands=[
                    dict(
                        type="init",
                        componentExecName="/wiki/edit",
                        initParams=dict(
                            page_path="root/jembe/suite",
                            form=dict(error=None, title="Jembe Suite Page"),
                        ),
                        mergeExistingParams=True,
                    ),
                    dict(
                        type="call",
                        componentExecName="/wiki/edit",
                        actionName="save",
                        args=list(),
                        kwargs=dict(),
                    ),
                ],
            )
        ),
        headers={"x-jembe": True},
    )
    assert r.status_code == 200
    json_response = json.loads(r.data)
    assert len(json_response) == 3
    assert json_response[0]["execName"] == "/wiki"
    assert json_response[0]["state"] == dict(mode="view")
    assert json_response[0]["dom"] == (
        """<html><head><template jmb-placeholder="/wiki/page_title"></template></head>"""
        """<body><template jmb-placeholder="/wiki/view"></template></body></html>"""
    )
    assert json_response[1]["execName"] == "/wiki/page_title"
    assert json_response[1]["state"] == dict(title="Jembe Suite Page")
    assert json_response[1]["dom"] == ("""<title>Jembe Suite Page</title>""")
    assert json_response[2]["execName"] == "/wiki/view"
    assert json_response[2]["state"] == dict(page_path="root/jembe/suite")
    assert json_response[2]["dom"] == (
        """<div><h1>Jembe Suite Page</h1><div></div></div>"""
    )
    # get wiki page via direct http get
    ######################################
    r = client.get("/wiki/view/root/jembe")
    assert r.status_code == 200
    assert r.data == (
        """<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN" "http://www.w3.org/TR/REC-html40/loose.dtd">"""
        "\n"
        """<html jmb-name="/wiki" jmb-data=\'{"actions":{},"changesUrl":true,"state":{"mode":"view"},"url":"/wiki"}\'>"""
        """<head><title jmb-name="/wiki/page_title" jmb-data=\'{"actions":{},"changesUrl":false,"state":{"title":"Jembe Page"},"url":"/wiki/page_title"}\'>Jembe Page</title></head>"""
        """<body>"""
        """<div jmb-name="/wiki/view" jmb-data=\'{"actions":{},"changesUrl":true,"state":{"page_path":"root/jembe"},"url":"/wiki/view/root/jembe"}\'>"""
        """<h1>Jembe Page</h1>"""
        """<div><a jmb-on:click="$jmb.set('page_path',\'root/jembe/suite\')">Jembe Suite Page</a><br></div>"""
        """</div>"""
        """</body></html>"""
    ).encode("utf-8")

    # TODO add back link, navigation menu etc in demo app
